"""CSSPageRule implements DOM Level 2 CSS CSSPageRule."""

__all__ = ['CSSPageRule']

import xml.dom
from itertools import chain

import cssutils

from . import cssrule
from .cssstyledeclaration import CSSStyleDeclaration
from .marginrule import MarginRule


class CSSPageRule(cssrule.CSSRuleRules):
    """
    The CSSPageRule interface represents a @page rule within a CSS style
    sheet. The @page rule is used to specify the dimensions, orientation,
    margins, etc. of a page box for paged media.

    Format::

        page :
               PAGE_SYM S* IDENT? pseudo_page? S*
               '{' S* [ declaration | margin ]?
               [ ';' S* [ declaration | margin ]? ]* '}' S*
               ;

        pseudo_page :
               ':' [ "left" | "right" | "first" ]
               ;

        margin :
               margin_sym S* '{' declaration [ ';' S* declaration? ]* '}' S*
               ;

        margin_sym :
               TOPLEFTCORNER_SYM |
               TOPLEFT_SYM |
               TOPCENTER_SYM |
               TOPRIGHT_SYM |
               TOPRIGHTCORNER_SYM |
               BOTTOMLEFTCORNER_SYM |
               BOTTOMLEFT_SYM |
               BOTTOMCENTER_SYM |
               BOTTOMRIGHT_SYM |
               BOTTOMRIGHTCORNER_SYM |
               LEFTTOP_SYM |
               LEFTMIDDLE_SYM |
               LEFTBOTTOM_SYM |
               RIGHTTOP_SYM |
               RIGHTMIDDLE_SYM |
               RIGHTBOTTOM_SYM
               ;

    `cssRules` contains a list of `MarginRule` objects.
    """

    def __init__(
        self,
        selectorText=None,
        style=None,
        parentRule=None,
        parentStyleSheet=None,
        readonly=False,
    ):
        """
        If readonly allows setting of properties in constructor only.

        :param selectorText:
            type string
        :param style:
            CSSStyleDeclaration for this CSSStyleRule
        """
        super().__init__(parentRule=parentRule, parentStyleSheet=parentStyleSheet)
        self._atkeyword = '@page'
        self._specificity = (0, 0, 0)

        tempseq = self._tempSeq()

        if selectorText:
            self.selectorText = selectorText
            tempseq.append(self.selectorText, 'selectorText')
        else:
            self._selectorText = self._tempSeq()

        if style:
            self.style = style
        else:
            self.style = CSSStyleDeclaration()

        tempseq.append(self.style, 'style')

        self._setSeq(tempseq)
        self._readonly = readonly

    def __repr__(self):
        return f"cssutils.css.{self.__class__.__name__}(selectorText={self.selectorText!r}, style={self.style.cssText!r})"

    def __str__(self):
        return (
            "<cssutils.css.%s object selectorText=%r specificity=%r "
            + "style=%r cssRules=%r at 0x%x>"
        ) % (
            self.__class__.__name__,
            self.selectorText,
            self.specificity,
            self.style.cssText,
            len(self.cssRules),
            id(self),
        )

    def __contains__(self, margin):
        """Check if margin is set in the rule."""
        return margin in list(self.keys())

    def keys(self):
        "Return list of all set margins (MarginRule)."
        return list(r.margin for r in self.cssRules)

    def __getitem__(self, margin):
        """Retrieve the style (of MarginRule)
        for `margin` (which must be normalized).
        """
        for r in self.cssRules:
            if r.margin == margin:
                return r.style

    def __setitem__(self, margin, style):
        """Set the style (of MarginRule)
        for `margin` (which must be normalized).
        """
        for i, r in enumerate(self.cssRules):
            if r.margin == margin:
                r.style = style
                return i
        else:
            return self.add(MarginRule(margin, style))

    def __delitem__(self, margin):
        """Delete the style (the MarginRule)
        for `margin` (which must be normalized).
        """
        for r in self.cssRules:
            if r.margin == margin:
                self.deleteRule(r)

    def __parseSelectorText(self, selectorText):  # noqa: C901
        """
        Parse `selectorText` which may also be a list of tokens
        and returns (selectorText, seq).

        see _setSelectorText for details
        """
        # for closures: must be a mutable
        new = {'wellformed': True, 'last-S': False, 'name': 0, 'first': 0, 'lr': 0}

        def _char(expected, seq, token, tokenizer=None):
            # pseudo_page, :left, :right or :first
            val = self._tokenvalue(token)
            if not new['last-S'] and expected in ['page', ': or EOF'] and ':' == val:
                try:
                    identtoken = next(tokenizer)
                except StopIteration:
                    self._log.error('CSSPageRule selectorText: No IDENT found.', token)
                else:
                    ival, ityp = self._tokenvalue(identtoken), self._type(identtoken)
                    if self._prods.IDENT != ityp:
                        self._log.error(
                            'CSSPageRule selectorText: Expected '
                            'IDENT but found: %r' % ival,
                            token,
                        )
                    else:
                        if ival not in ('first', 'left', 'right'):
                            self._log.warn(
                                'CSSPageRule: Unknown @page '
                                'selector: %r' % (':' + ival,),
                                neverraise=True,
                            )
                        if ival == 'first':
                            new['first'] = 1
                        else:
                            new['lr'] = 1
                        seq.append(val + ival, 'pseudo')
                        return 'EOF'
                return expected
            else:
                new['wellformed'] = False
                self._log.error(
                    'CSSPageRule selectorText: Unexpected CHAR: %r' % val, token
                )
                return expected

        def S(expected, seq, token, tokenizer=None):
            "Does not raise if EOF is found."
            if expected == ': or EOF':
                # pseudo must directly follow IDENT if given
                new['last-S'] = True
            return expected

        def IDENT(expected, seq, token, tokenizer=None):
            """ """
            val = self._tokenvalue(token)
            if 'page' == expected:
                if self._normalize(val) == 'auto':
                    self._log.error(
                        'CSSPageRule selectorText: Invalid pagename.', token
                    )
                else:
                    new['name'] = 1
                    seq.append(val, 'IDENT')

                return ': or EOF'
            else:
                new['wellformed'] = False
                self._log.error(
                    'CSSPageRule selectorText: Unexpected IDENT: ' '%r' % val, token
                )
                return expected

        def COMMENT(expected, seq, token, tokenizer=None):
            "Does not raise if EOF is found."
            seq.append(cssutils.css.CSSComment([token]), 'COMMENT')
            return expected

        newseq = self._tempSeq()
        wellformed, expected = self._parse(
            expected='page',
            seq=newseq,
            tokenizer=self._tokenize2(selectorText),
            productions={'CHAR': _char, 'IDENT': IDENT, 'COMMENT': COMMENT, 'S': S},
            new=new,
        )
        wellformed = wellformed and new['wellformed']

        # post conditions
        if expected == 'ident':
            self._log.error(
                'CSSPageRule selectorText: No valid selector: %r'
                % self._valuestr(selectorText)
            )

        return wellformed, newseq, (new['name'], new['first'], new['lr'])

    def __parseMarginAndStyle(self, tokens):
        "tokens is a list, no generator (yet)"
        g = iter(tokens)
        styletokens = []

        # new rules until parse done
        cssRules = []

        for token in g:
            if (
                token[0] == 'ATKEYWORD'
                and self._normalize(token[1]) in MarginRule.margins
            ):
                # MarginRule
                m = MarginRule(parentRule=self, parentStyleSheet=self.parentStyleSheet)
                m.cssText = chain([token], g)

                # merge if margin set more than once
                for r in cssRules:
                    if r.margin == m.margin:
                        for p in m.style:
                            r.style.setProperty(p, replace=False)
                        break
                else:
                    cssRules.append(m)

                continue

            # TODO: Properties?
            styletokens.append(token)

        return cssRules, styletokens

    def _getCssText(self):
        """Return serialized property cssText."""
        return cssutils.ser.do_CSSPageRule(self)

    def _setCssText(self, cssText):
        """
        :exceptions:
            - :exc:`~xml.dom.SyntaxErr`:
              Raised if the specified CSS string value has a syntax error and
              is unparsable.
            - :exc:`~xml.dom.InvalidModificationErr`:
              Raised if the specified CSS string value represents a different
              type of rule than the current one.
            - :exc:`~xml.dom.HierarchyRequestErr`:
              Raised if the rule cannot be inserted at this point in the
              style sheet.
            - :exc:`~xml.dom.NoModificationAllowedErr`:
              Raised if the rule is readonly.
        """
        super()._setCssText(cssText)

        tokenizer = self._tokenize2(cssText)
        if self._type(self._nexttoken(tokenizer)) != self._prods.PAGE_SYM:
            self._log.error(
                'CSSPageRule: No CSSPageRule found: %s' % self._valuestr(cssText),
                error=xml.dom.InvalidModificationErr,
            )
        else:
            newStyle = CSSStyleDeclaration(parentRule=self)
            ok = True

            selectortokens, startbrace = self._tokensupto2(
                tokenizer, blockstartonly=True, separateEnd=True
            )
            styletokens, braceorEOFtoken = self._tokensupto2(
                tokenizer, blockendonly=True, separateEnd=True
            )
            nonetoken = self._nexttoken(tokenizer)
            if self._tokenvalue(startbrace) != '{':
                ok = False
                self._log.error(
                    'CSSPageRule: No start { of style declaration '
                    'found: %r' % self._valuestr(cssText),
                    startbrace,
                )
            elif nonetoken:
                ok = False
                self._log.error('CSSPageRule: Trailing content found.', token=nonetoken)

            selok, newselseq, specificity = self.__parseSelectorText(selectortokens)
            ok = ok and selok

            val, type_ = self._tokenvalue(braceorEOFtoken), self._type(braceorEOFtoken)

            if val != '}' and type_ != 'EOF':
                ok = False
                self._log.error(
                    'CSSPageRule: No "}" after style declaration found: %r'
                    % self._valuestr(cssText)
                )
            else:
                if 'EOF' == type_:
                    # add again as style needs it
                    styletokens.append(braceorEOFtoken)

                # filter pagemargin rules out first
                cssRules, styletokens = self.__parseMarginAndStyle(styletokens)

                # SET, may raise:
                newStyle.cssText = styletokens

            if ok:
                self._selectorText = newselseq
                self._specificity = specificity
                self.style = newStyle
                self.cssRules = cssutils.css.CSSRuleList()
                for r in cssRules:
                    self.cssRules.append(r)

    cssText = property(
        _getCssText,
        _setCssText,
        doc="(DOM) The parsable textual representation of this rule.",
    )

    def _getSelectorText(self):
        """Wrapper for cssutils Selector object."""
        return cssutils.ser.do_CSSPageRuleSelector(self._selectorText)

    def _setSelectorText(self, selectorText):
        """Wrapper for cssutils Selector object.

        :param selectorText:
            DOM String, in CSS 2.1 one of

            - :first
            - :left
            - :right
            - empty

        :exceptions:
            - :exc:`~xml.dom.SyntaxErr`:
              Raised if the specified CSS string value has a syntax error
              and is unparsable.
            - :exc:`~xml.dom.NoModificationAllowedErr`:
              Raised if this rule is readonly.
        """
        self._checkReadonly()

        # may raise SYNTAX_ERR
        wellformed, newseq, specificity = self.__parseSelectorText(selectorText)
        if wellformed:
            self._selectorText = newseq
            self._specificity = specificity

    selectorText = property(
        _getSelectorText,
        _setSelectorText,
        doc="(DOM) The parsable textual representation of "
        "the page selector for the rule.",
    )

    def _setStyle(self, style):
        """
        :param style:
            a CSSStyleDeclaration or string
        """
        self._checkReadonly()
        if isinstance(style, str):
            self._style = CSSStyleDeclaration(cssText=style, parentRule=self)
        else:
            style._parentRule = self
            self._style = style

    style = property(
        lambda self: self._style,
        _setStyle,
        doc="(DOM) The declaration-block of this rule set, "
        "a :class:`~cssutils.css.CSSStyleDeclaration`.",
    )

    def insertRule(self, rule, index=None):
        """Implements base ``insertRule``."""
        rule, index = self._prepareInsertRule(rule, index)

        if rule is False or rule is True:
            # done or error
            return

        # check hierarchy
        if (
            isinstance(rule, cssutils.css.CSSCharsetRule)
            or isinstance(rule, cssutils.css.CSSFontFaceRule)
            or isinstance(rule, cssutils.css.CSSImportRule)
            or isinstance(rule, cssutils.css.CSSNamespaceRule)
            or isinstance(rule, CSSPageRule)
            or isinstance(rule, cssutils.css.CSSMediaRule)
        ):
            self._log.error(
                '%s: This type of rule is not allowed here: %s'
                % (self.__class__.__name__, rule.cssText),
                error=xml.dom.HierarchyRequestErr,
            )
            return

        return self._finishInsertRule(rule, index)

    specificity = property(
        lambda self: self._specificity,
        doc="""Specificity of this page rule (READONLY).
Tuple of (f, g, h) where:

 - if the page selector has a named page, f=1; else f=0
 - if the page selector has a ':first' pseudo-class, g=1; else g=0
 - if the page selector has a ':left' or ':right' pseudo-class, h=1; else h=0
""",
    )

    type = property(
        lambda self: self.PAGE_RULE,
        doc="The type of this rule, as defined by a CSSRule " "type constant.",
    )

    # constant but needed:
    wellformed = property(lambda self: True)



"""Testcases for cssutils.stylesheets.MediaQuery"""

import xml.dom

import pytest

import cssutils.stylesheets

from . import basetest


class TestMediaQuery(basetest.BaseTestCase):
    def setup_method(self):
        self.r = cssutils.stylesheets.MediaQuery()

    def test_mediaText(self):
        "MediaQuery.mediaText"
        tests = {
            'all': None,
            'braille': None,
            'embossed': None,
            'handheld': None,
            'print': None,
            'projection': None,
            'screen': None,
            'speech': None,
            'tty': None,
            'tv': None,
            'ALL': None,
            'a\\ll': None,
            'not tv': None,
            'n\\ot t\\v': None,
            'only tv': None,
            '\\only \\tv': None,
            'PRINT': None,
            'NOT PRINT': None,
            'ONLY PRINT': None,
            'tv and (color)': None,
            'not tv and (color)': None,
            'only tv and (color)': None,
            'print and(color)': 'print and (color)',
        }
        self.do_equal_r(tests, att='mediaText')

        tests = {
            '': xml.dom.SyntaxErr,
            'two values': xml.dom.SyntaxErr,
            'or even three': xml.dom.SyntaxErr,
            'aural': xml.dom.SyntaxErr,  # a dimension
            '3d': xml.dom.SyntaxErr,  # a dimension
        }
        self.do_raise_r(tests, att='_setMediaText')

    def test_mediaType(self):
        "MediaQuery.mediaType"
        mq = cssutils.stylesheets.MediaQuery()

        assert '' == mq.mediaText

        for mt in cssutils.stylesheets.MediaQuery.MEDIA_TYPES:
            mq.mediaType = mt
            assert mq.mediaType == mt
            mq.mediaType = mt.upper()
            assert mq.mediaType == mt.upper()

        mt = '3D-UNKOwn-MEDIAtype0123'
        # mq.mediaType = mt
        with pytest.raises(xml.dom.SyntaxErr):
            mq._setMediaType(mt)
        # self.assertRaises(xml.dom.InvalidCharacterErr, mq._setMediaType, mt)

    def test_comments(self):
        "MediaQuery.mediaText comments"
        tests = {
            'all': None,
            'print': None,
            'not print': None,
            'only print': None,
            'print and (color)': None,
            'print and (color) and (width)': None,
            'print and (color: 2)': None,
            'print and (min-width: 100px)': None,
            'print and (min-width: 100px) and (color: red)': None,
            'not print and (min-width: 100px)': None,
            'only print and (min-width: 100px)': None,
            '/*1*/ tv /*2*/': None,
            '/*0*/ only /*1*/ tv /*2*/': None,
            '/*0* /not /*1*/ tv /*2*/': None,
            '/*x*/ only /*x*/ print /*x*/ and /*x*/ (/*x*/ '
            'min-width /*x*/: /*x*/ 100px /*x*/)': None,
            'print and/*1*/(color)': 'print and /*1*/ (color)',
        }
        self.do_equal_r(tests, att='mediaText')

    def test_reprANDstr(self):
        "MediaQuery.__repr__(), .__str__()"
        mediaText = 'tv and (color)'
        s = cssutils.stylesheets.MediaQuery(mediaText=mediaText)
        assert mediaText in str(s)
        s2 = eval(repr(s))
        assert mediaText == s2.mediaText
        assert isinstance(s2, s.__class__)



"""Testcases for cssutils.css.CSSRule"""

import xml.dom

import pytest

import cssutils.css

from . import basetest


class TestCSSRule(basetest.BaseTestCase):
    """
    base class for all CSSRule subclass tests

    overwrite setUp with the appriopriate values, will be used in
    test_init and test_readonly
    overwrite all tests as you please, use::

        super(CLASSNAME, self).test_TESTNAME(params)

    to use the base class tests too
    """

    def setup_method(self):
        """
        self.r is the rule
        self.rRO the readonly rule
        relf.r_type the type as defined in CSSRule
        """
        self._setup_rule()
        self.sheet = cssutils.css.CSSStyleSheet()

    def _setup_rule(self):
        self.r = cssutils.css.CSSRule()
        self.rRO = cssutils.css.CSSRule()
        self.rRO._readonly = True  # must be set here!
        self.r_type = cssutils.css.CSSRule.UNKNOWN_RULE
        self.r_typeString = 'UNKNOWN_RULE'

    def test_init(self):
        "CSSRule.type and init"
        assert self.r_type == self.r.type
        assert self.r_typeString == self.r.typeString
        assert '' == self.r.cssText
        assert self.r.parentRule is None
        assert self.r.parentStyleSheet is None

    def test_parentRule_parentStyleSheet_type(self):  # noqa: C901
        "CSSRule.parentRule .parentStyleSheet .type"
        rules = [
            ('@charset "ascii";', cssutils.css.CSSRule.CHARSET_RULE),
            ('@import "x";', cssutils.css.CSSRule.IMPORT_RULE),
            ('@namespace "x";', cssutils.css.CSSRule.NAMESPACE_RULE),
            ('@font-face { src: url(x) }', cssutils.css.CSSRule.FONT_FACE_RULE),
            (
                '''@media all {
                    @x;
                    a { color: red }
                    /* c  */
                }''',
                cssutils.css.CSSRule.MEDIA_RULE,
            ),
            ('@page :left { color: red }', cssutils.css.CSSRule.PAGE_RULE),
            ('@unknown;', cssutils.css.CSSRule.UNKNOWN_RULE),
            ('b { left: 0 }', cssutils.css.CSSRule.STYLE_RULE),
            ('/*1*/', cssutils.css.CSSRule.COMMENT),  # must be last for add test
        ]
        mrt = [
            cssutils.css.CSSRule.UNKNOWN_RULE,
            cssutils.css.CSSRule.STYLE_RULE,
            cssutils.css.CSSRule.COMMENT,
        ]

        def test(s):
            for i, rule in enumerate(s):
                assert rule.parentRule is None
                assert rule.parentStyleSheet == s
                # self.assertEqual(rule.type, rules[i][1])
                if rule.MEDIA_RULE == rule.type:
                    for j, r in enumerate(rule):
                        assert r.parentRule == rule
                        assert r.parentStyleSheet == s
                        assert r.type == mrt[j]

                if i == 0:  # check encoding
                    assert 'ascii' == s.encoding
                elif i == 2:  # check namespaces
                    assert 'x' == s.namespaces['']

        cssText = ''.join(r[0] for r in rules)
        # parsing
        s = cssutils.parseString(cssText)
        test(s)
        # sheet.cssText
        s = cssutils.css.CSSStyleSheet()
        s.cssText = cssText
        test(s)
        # sheet.add CSS
        s = cssutils.css.CSSStyleSheet()
        for css, _ in rules:
            s.add(css)
        test(s)
        # sheet.insertRule CSS
        s = cssutils.css.CSSStyleSheet()
        for css, _ in rules:
            s.insertRule(css)
        test(s)

        types = [
            cssutils.css.CSSCharsetRule,
            cssutils.css.CSSImportRule,
            cssutils.css.CSSNamespaceRule,
            cssutils.css.CSSFontFaceRule,
            cssutils.css.CSSMediaRule,
            cssutils.css.CSSPageRule,
            cssutils.css.CSSUnknownRule,
            cssutils.css.CSSStyleRule,
            cssutils.css.CSSComment,
        ]
        # sheet.add CSSRule
        s = cssutils.css.CSSStyleSheet()
        for i, (css, _) in enumerate(rules):
            rule = types[i]()
            rule.cssText = css
            s.add(rule)
        test(s)
        # sheet.insertRule CSSRule
        s = cssutils.css.CSSStyleSheet()
        for i, (css, _) in enumerate(rules):
            rule = types[i]()
            rule.cssText = css
            s.insertRule(rule)
        test(s)

    def test_CSSMediaRule_cssRules_parentRule_parentStyleSheet_type(self):
        "CSSMediaRule.cssRules.parentRule .parentStyleSheet .type"
        rules = [
            ('b { left: 0 }', cssutils.css.CSSRule.STYLE_RULE),
            ('/*1*/', cssutils.css.CSSRule.COMMENT),
            ('@x;', cssutils.css.CSSRule.UNKNOWN_RULE),
        ]

        def test(s):
            mr = s.cssRules[0]
            for i, rule in enumerate(mr):
                assert rule.parentRule == mr
                assert rule.parentStyleSheet == s
                assert rule.parentStyleSheet == mr.parentStyleSheet
                assert rule.type == rules[i][1]

        cssText = '@media all { %s }' % ''.join(r[0] for r in rules)
        # parsing
        s = cssutils.parseString(cssText)
        test(s)
        # sheet.cssText
        s = cssutils.css.CSSStyleSheet()
        s.cssText = cssText
        test(s)

        def getMediaSheet():
            s = cssutils.css.CSSStyleSheet()
            s.cssText = '@media all {}'
            return s, s.cssRules[0]

        # sheet.add CSS
        s, mr = getMediaSheet()
        for css, _ in rules:
            mr.add(css)
        test(s)
        # sheet.insertRule CSS
        s, mr = getMediaSheet()
        for css, _ in rules:
            mr.insertRule(css)
        test(s)

        types = [
            cssutils.css.CSSStyleRule,
            cssutils.css.CSSComment,
            cssutils.css.CSSUnknownRule,
        ]
        # sheet.add CSSRule
        s, mr = getMediaSheet()
        for i, (css, _) in enumerate(rules):
            rule = types[i]()
            rule.cssText = css
            mr.add(rule)
        test(s)
        # sheet.insertRule CSSRule
        s, mr = getMediaSheet()
        for i, (css, _) in enumerate(rules):
            rule = types[i]()
            rule.cssText = css
            mr.insertRule(rule)
        test(s)

    def test_readonly(self):
        "CSSRule readonly"
        self.rRO = cssutils.css.CSSRule()
        self.rRO._readonly = True
        assert self.rRO._readonly
        assert '' == self.rRO.cssText
        with pytest.raises(xml.dom.NoModificationAllowedErr):
            self.rRO._setCssText('x')
        assert '' == self.rRO.cssText

    def _test_InvalidModificationErr(self, startwithspace):
        """
        CSSRule.cssText InvalidModificationErr

        called by subclasses

        startwithspace

        for test starting with this not the test but " test" is tested
        e.g. " @page {}"
        exception is the style rule test
        """
        tests = (
            '',
            '/* comment */',
            '@charset "utf-8";',
            '@font-face {}',
            '@import url(x);',
            '@media all {}',
            '@namespace "x";' '@page {}',
            '@unknown;',
            '@variables;',
            # TODO:
            # u'@top-left {}'
            'a style rule {}',
        )
        for test in tests:
            if startwithspace in ('a style rule',) and test in (
                '/* comment */',
                'a style rule {}',
            ):
                continue

            if test.startswith(startwithspace):
                test = ' %s' % test

            with pytest.raises(xml.dom.InvalidModificationErr):
                self.r._setCssText(test)

        # check that type is readonly
        with pytest.raises(AttributeError):
            self.r.__setattr__('parentRule', None)
        with pytest.raises(AttributeError):
            self.r.__setattr__('parentStyleSheet', None)
        with pytest.raises(AttributeError):
            self.r.__setattr__('type', 1)
        with pytest.raises(AttributeError):
            self.r.__setattr__('typeString', "")



"""CSSVariables implements (and only partly) experimental
`CSS Variables <http://disruptive-innovations.com/zoo/cssvariables/>`_
"""

__all__ = ['CSSVariablesRule']

import xml.dom

import cssutils

from . import cssrule
from .cssvariablesdeclaration import CSSVariablesDeclaration


class CSSVariablesRule(cssrule.CSSRule):
    """
    The CSSVariablesRule interface represents a @variables rule within a CSS
    style sheet. The @variables rule is used to specify variables.

    cssutils uses a :class:`~cssutils.css.CSSVariablesDeclaration`  to
    represent the variables.

    Format::

        variables
            VARIABLES_SYM S* medium [ COMMA S* medium ]* LBRACE S*
            variableset* '}' S*
            ;

    for variableset see :class:`cssutils.css.CSSVariablesDeclaration`

    **Media are not implemented. Reason is that cssutils is using CSS
    variables in a kind of preprocessing and therefor no media information
    is available at this stage. For now do not use media!**

    Example::

        @variables {
          CorporateLogoBGColor: #fe8d12;
        }

        div.logoContainer {
          background-color: var(CorporateLogoBGColor);
        }
    """

    def __init__(
        self,
        mediaText=None,
        variables=None,
        parentRule=None,
        parentStyleSheet=None,
        readonly=False,
    ):
        """
        If readonly allows setting of properties in constructor only.
        """
        super().__init__(parentRule=parentRule, parentStyleSheet=parentStyleSheet)
        self._atkeyword = '@variables'

        # dummy
        self._media = cssutils.stylesheets.MediaList(mediaText, readonly=readonly)

        if variables:
            self.variables = variables
        else:
            self.variables = CSSVariablesDeclaration(parentRule=self)

        self._readonly = readonly

    def __repr__(self):
        return f"cssutils.css.{self.__class__.__name__}(mediaText={self._media.mediaText!r}, variables={self.variables.cssText!r})"

    def __str__(self):
        return (
            "<cssutils.css.%s object mediaText=%r variables=%r valid=%r "
            "at 0x%x>"
            % (
                self.__class__.__name__,
                self._media.mediaText,
                self.variables.cssText,
                self.valid,
                id(self),
            )
        )

    def _getCssText(self):
        """Return serialized property cssText."""
        return cssutils.ser.do_CSSVariablesRule(self)

    def _setCssText(self, cssText):
        """
        :exceptions:
            - :exc:`~xml.dom.SyntaxErr`:
              Raised if the specified CSS string value has a syntax error and
              is unparsable.
            - :exc:`~xml.dom.InvalidModificationErr`:
              Raised if the specified CSS string value represents a different
              type of rule than the current one.
            - :exc:`~xml.dom.HierarchyRequestErr`:
              Raised if the rule cannot be inserted at this point in the
              style sheet.
            - :exc:`~xml.dom.NoModificationAllowedErr`:
              Raised if the rule is readonly.

        Format::

            variables
            : VARIABLES_SYM S* medium [ COMMA S* medium ]* LBRACE S*
              variableset* '}' S*
            ;

            variableset
            : LBRACE S* vardeclaration [ ';' S* vardeclaration ]* '}' S*
            ;
        """
        super()._setCssText(cssText)

        tokenizer = self._tokenize2(cssText)
        attoken = self._nexttoken(tokenizer, None)
        if self._type(attoken) != self._prods.VARIABLES_SYM:
            self._log.error(
                'CSSVariablesRule: No CSSVariablesRule found: %s'
                % self._valuestr(cssText),
                error=xml.dom.InvalidModificationErr,
            )
        else:
            newVariables = CSSVariablesDeclaration(parentRule=self)
            ok = True

            beforetokens, brace = self._tokensupto2(
                tokenizer, blockstartonly=True, separateEnd=True
            )
            if self._tokenvalue(brace) != '{':
                ok = False
                self._log.error(
                    'CSSVariablesRule: No start { of variable '
                    'declaration found: %r' % self._valuestr(cssText),
                    brace,
                )

            # parse stuff before { which should be comments and S only
            new = {'wellformed': True}
            newseq = self._tempSeq()  # []

            beforewellformed, expected = self._parse(
                expected=':',
                seq=newseq,
                tokenizer=self._tokenize2(beforetokens),
                productions={},
            )
            ok = ok and beforewellformed and new['wellformed']

            variablestokens, braceorEOFtoken = self._tokensupto2(
                tokenizer, blockendonly=True, separateEnd=True
            )

            val, type_ = self._tokenvalue(braceorEOFtoken), self._type(braceorEOFtoken)
            if val != '}' and type_ != 'EOF':
                ok = False
                self._log.error(
                    'CSSVariablesRule: No "}" after variables '
                    'declaration found: %r' % self._valuestr(cssText)
                )

            nonetoken = self._nexttoken(tokenizer)
            if nonetoken:
                ok = False
                self._log.error(
                    'CSSVariablesRule: Trailing content found.', token=nonetoken
                )

            if 'EOF' == type_:
                # add again as variables needs it
                variablestokens.append(braceorEOFtoken)
            # SET but may raise:
            newVariables.cssText = variablestokens

            if ok:
                # contains probably comments only upto {
                self._setSeq(newseq)
                self.variables = newVariables

    cssText = property(
        _getCssText,
        _setCssText,
        doc="(DOM) The parsable textual representation of this " "rule.",
    )

    media = property(
        doc="NOT IMPLEMENTED! As cssutils resolves variables "
        "during serializing media information is lost."
    )

    def _setVariables(self, variables):
        """
        :param variables:
            a CSSVariablesDeclaration or string
        """
        self._checkReadonly()
        if isinstance(variables, str):
            self._variables = CSSVariablesDeclaration(
                cssText=variables, parentRule=self
            )
        else:
            variables._parentRule = self
            self._variables = variables

    variables = property(
        lambda self: self._variables,
        _setVariables,
        doc="(DOM) The variables of this rule set, a "
        ":class:`cssutils.css.CSSVariablesDeclaration`.",
    )

    type = property(
        lambda self: self.VARIABLES_RULE,
        doc="The type of this rule, as defined by a CSSRule " "type constant.",
    )

    valid = property(lambda self: True, doc='NOT IMPLEMTED REALLY (TODO)')

    # constant but needed:
    wellformed = property(lambda self: True)



"""Testcases for cssutils.codec"""

import codecs
import io

import pytest

from cssutils import codec

iostream = io.BytesIO


class Queue:
    """
    queue: write bytes at one end, read bytes from the other end
    """

    def __init__(self):
        self._buffer = b""

    def write(self, chars):
        # TODO ???
        if isinstance(chars, str):
            chars = chars.encode()
        elif isinstance(chars, int):
            chars = bytes([chars])

        self._buffer += chars

    def read(self, size=-1):
        if size < 0:
            s = self._buffer
            self._buffer = b""
            return s
        else:
            s = self._buffer[:size]
            self._buffer = self._buffer[size:]
            return s


class TestCodec:
    def test_detectencoding_str(self):
        "codec.detectencoding_str()"
        assert codec.detectencoding_str(b'') == (None, False)
        assert codec.detectencoding_str(b'\xef') == (None, False)
        assert codec.detectencoding_str('\xef\x33'.encode()) == ("utf-8", False)
        assert codec.detectencoding_str('\xc3\xaf3'.encode()) == ("utf-8", False)
        assert codec.detectencoding_str(b'\xef\xbb') == (None, False)
        assert codec.detectencoding_str('\xef\xbb\x33'.encode()) == (
            "utf-8",
            False,
        )
        assert codec.detectencoding_str('\xef\xbb\xbf'.encode("utf-8-sig")) == (
            "utf-8-sig",
            True,
        )
        assert codec.detectencoding_str(b'\xff') == (None, False)
        assert codec.detectencoding_str('\xff\x33'.encode()) == ("utf-8", False)
        assert codec.detectencoding_str(b'\xff\xfe') == (None, False)
        assert codec.detectencoding_str('\xff\xfe\x33'.encode("utf-16")) == (
            "utf-16",
            True,
        )
        assert codec.detectencoding_str(b'\xff\xfe\x00') == (
            None,
            False,
        )
        assert codec.detectencoding_str('\xff\xfe\x00\x33'.encode("utf-16")) == (
            "utf-16",
            True,
        )
        assert codec.detectencoding_str('\xff\xfe\x00\x00'.encode("utf-32")) == (
            "utf-32",
            True,
        )
        assert codec.detectencoding_str(b'\x00') == (None, False)
        assert codec.detectencoding_str(b'\x00\x33') == ("utf-8", False)
        assert codec.detectencoding_str(b'\x00\x00') == (None, False)
        assert codec.detectencoding_str(b'\x00\x00\x33') == ("utf-8", False)
        assert codec.detectencoding_str(b'\x00\x00\xfe') == (
            None,
            False,
        )
        assert codec.detectencoding_str(b'\x00\x00\x00\x33') == ("utf-8", False)
        assert codec.detectencoding_str(b'\x00\x00\x00@') == (
            "utf-32-be",
            False,
        )
        assert codec.detectencoding_str('\x00\x00\xfe\xff'.encode('utf-32')) == (
            "utf-32",
            True,
        )
        assert codec.detectencoding_str(b'@') == (None, False)
        assert codec.detectencoding_str(b'@\x33') == ("utf-8", False)
        assert codec.detectencoding_str(b'@\x00') == (None, False)
        assert codec.detectencoding_str(b'@\x00\x33') == ("utf-8", False)
        assert codec.detectencoding_str(b'@\x00\x00') == (None, False)
        assert codec.detectencoding_str(b'@\x00\x00\x33') == ("utf-8", False)
        assert codec.detectencoding_str(b'@\x00\x00\x00') == (
            "utf-32-le",
            False,
        )
        assert codec.detectencoding_str(b'@c') == (None, False)
        assert codec.detectencoding_str(b'@ch') == (None, False)
        assert codec.detectencoding_str(b'@cha') == (None, False)
        assert codec.detectencoding_str(b'@char') == (None, False)
        assert codec.detectencoding_str(b'@chars') == (None, False)
        assert codec.detectencoding_str(b'@charse') == (None, False)
        assert codec.detectencoding_str(b'@charset') == (None, False)
        assert codec.detectencoding_str(b'@charset ') == (None, False)
        assert codec.detectencoding_str(b'@charset "') == (None, False)
        assert codec.detectencoding_str(b'@charset "x') == (None, False)
        assert codec.detectencoding_str(b'@charset ""') == ("", True)
        assert codec.detectencoding_str(b'@charset "x"') == ("x", True)
        assert codec.detectencoding_str(b"@", False) == (None, False)
        assert codec.detectencoding_str(b"@", True) == ("utf-8", False)
        assert codec.detectencoding_str(b"@c", False) == (None, False)
        assert codec.detectencoding_str(b"@c", True) == ("utf-8", False)

    def test_detectencoding_unicode(self):
        "codec.detectencoding_unicode()"
        # Unicode version (only parses the header)
        assert codec.detectencoding_unicode('@charset "x') == (None, False)
        assert codec.detectencoding_unicode('a {}') == ("utf-8", False)
        assert codec.detectencoding_unicode('@charset "x', True) == (None, False)
        assert codec.detectencoding_unicode('@charset "x"') == ("x", True)

    def test_fixencoding(self):
        "codec._fixencoding()"
        s = '@charset "'
        assert codec._fixencoding(s, "utf-8") is None

        s = '@charset "x'
        assert codec._fixencoding(s, "utf-8") is None

        s = '@charset "x'
        assert codec._fixencoding(s, "utf-8", True) == s

        s = '@charset x'
        assert codec._fixencoding(s, "utf-8") == s

        s = '@charset "x"'
        assert codec._fixencoding(s, "utf-8") == s.replace('"x"', '"utf-8"')

    def test_decoder(self):  # noqa: C901
        "codecs.decoder"

        def checkauto(encoding, input='@charset "x";g\xfcrk\u20ac{}'):
            outputencoding = encoding
            if outputencoding == "utf-8-sig":
                outputencoding = "utf-8"
            # Check stateless decoder with encoding autodetection
            d = codecs.getdecoder("css")
            assert d(input.encode(encoding))[0] == input.replace(
                '"x"', '"%s"' % outputencoding
            )

            # Check stateless decoder with specified encoding
            assert d(input.encode(encoding), encoding=encoding)[0] == input.replace(
                '"x"', '"%s"' % outputencoding
            )

            if hasattr(codec, "getincrementaldecoder"):
                # Check incremental decoder with encoding autodetection
                id = codecs.getincrementaldecoder("css")()
                assert "".join(id.iterdecode(input.encode(encoding))) == input.replace(
                    '"x"', '"%s"' % outputencoding
                )

                # Check incremental decoder with specified encoding
                id = codecs.getincrementaldecoder("css")(encoding=encoding)
                assert "".join(id.iterdecode(input.encode(encoding))) == input.replace(
                    '"x"', '"%s"' % outputencoding
                )

            # Check stream reader with encoding autodetection
            q = Queue()
            sr = codecs.getreader("css")(q)
            result = []
            # TODO: py3 only???
            for c in input.encode(encoding):
                q.write(c)
                result.append(sr.read())
            assert "".join(result) == input.replace('"x"', '"%s"' % outputencoding)

            # Check stream reader with specified encoding
            q = Queue()
            sr = codecs.getreader("css")(q, encoding=encoding)
            result = []
            for c in input.encode(encoding):
                q.write(c)
                result.append(sr.read())
            assert "".join(result) == input.replace('"x"', '"%s"' % outputencoding)

        # Autodetectable encodings
        checkauto("utf-8-sig")
        checkauto("utf-16")
        checkauto("utf-16-le")
        checkauto("utf-16-be")
        checkauto("utf-32")
        checkauto("utf-32-le")
        checkauto("utf-32-be")

        def checkdecl(encoding, input='@charset "%s";g\xfcrk{}'):
            # Check stateless decoder with encoding autodetection
            d = codecs.getdecoder("css")
            input = input % encoding
            outputencoding = encoding
            if outputencoding == "utf-8-sig":
                outputencoding = "utf-8"
            assert d(input.encode(encoding))[0] == input

            # Check stateless decoder with specified encoding
            assert d(input.encode(encoding), encoding=encoding)[0] == input

            if hasattr(codec, "getincrementaldecoder"):
                # Check incremental decoder with encoding autodetection
                id = codecs.getincrementaldecoder("css")()
                assert "".join(id.iterdecode(input.encode(encoding))) == input

                # Check incremental decoder with specified encoding
                id = codecs.getincrementaldecoder("css")(encoding)
                assert "".join(id.iterdecode(input.encode(encoding))) == input

            # Check stream reader with encoding autodetection
            q = Queue()
            sr = codecs.getreader("css")(q)
            result = []
            for c in input.encode(encoding):
                q.write(c)
                result.append(sr.read())
            assert "".join(result) == input

            # Check stream reader with specified encoding
            q = Queue()
            sr = codecs.getreader("css")(q, encoding=encoding)
            result = []
            for c in input.encode(encoding):
                q.write(c)
                result.append(sr.read())
            assert "".join(result) == input

        # Use correct declaration
        checkdecl("utf-8")
        checkdecl("iso-8859-1", '@charset "%s";g\xfcrk')
        checkdecl("iso-8859-15")
        checkdecl("cp1252")

        # No recursion
        with pytest.raises(ValueError):
            b'@charset "css";div{}'.decode("css")

    def test_encoder(self):
        "codec.encoder"

        def check(encoding, input='@charset "x";g\xfcrk\u20ac{}'):
            outputencoding = encoding
            if outputencoding == "utf-8-sig":
                outputencoding = "utf-8"

            # Check stateless encoder with encoding autodetection
            e = codecs.getencoder("css")
            inputdecl = input.replace('"x"', '"%s"' % encoding)
            outputdecl = input.replace('"x"', '"%s"' % outputencoding)
            assert e(inputdecl)[0].decode(encoding) == outputdecl

            # Check stateless encoder with specified encoding
            assert e(input, encoding=encoding)[0].decode(encoding) == outputdecl

            if hasattr(codec, "getincrementalencoder"):
                # Check incremental encoder with encoding autodetection
                ie = codecs.getincrementalencoder("css")()
                assert "".join(ie.iterencode(inputdecl)).decode(encoding) == outputdecl

                # Check incremental encoder with specified encoding
                ie = codecs.getincrementalencoder("css")(encoding=encoding)
                assert "".join(ie.iterencode(input)).decode(encoding) == outputdecl

            # Check stream writer with encoding autodetection
            q = Queue()
            sw = codecs.getwriter("css")(q)
            for c in inputdecl:  # .encode(outputencoding): # TODO: .encode()???
                sw.write(c)
            assert q.read().decode(encoding) == input.replace(
                '"x"', '"%s"' % outputencoding
            )

            # Check stream writer with specified encoding
            q = Queue()
            sw = codecs.getwriter("css")(q, encoding=encoding)
            for c in input:
                sw.write(c)
            assert q.read().decode(encoding) == input.replace(
                '"x"', '"%s"' % outputencoding
            )

        # Autodetectable encodings
        check("utf-8-sig")
        check("utf-16")
        check("utf-16-le")
        check("utf-16-be")
        check("utf-32")
        check("utf-32-le")
        check("utf-32-be")
        check("utf-8")
        check("iso-8859-1", '@charset "x";g\xfcrk{}')
        check("iso-8859-15")
        check("cp1252")

        # No recursion
        with pytest.raises(ValueError):
            '@charset "css";div{}'.encode("css")

    def test_decode_force(self):
        "codec.decode (force)"
        info = codecs.lookup("css")

        def decodeall(input, **kwargs):
            # Py 2.5: info.decode('@charset "utf-8"; x')
            return info[1](input, **kwargs)[0]

        def incdecode(input, **kwargs):
            decoder = info.incrementaldecoder(**kwargs)
            return decoder.decode(input)

        def streamdecode(input, **kwargs):
            stream = iostream(input)  # py3 .decode('utf-8') but still error?!
            reader = info.streamreader(stream, **kwargs)
            return reader.read()

        for d in (decodeall, incdecode, streamdecode):
            # input = '@charset "utf-8"; \xc3\xbf'
            # output = u'@charset "utf-8"; \xff'
            # self.assertEqual(d(input), output)
            #
            # input = '@charset "utf-8"; \xc3\xbf'
            # output = u'@charset "iso-8859-1"; \xc3\xbf'
            # self.assertEqual(d(input, encoding="iso-8859-1", force=True), output)
            #
            # input = '\xc3\xbf'
            # output = u'\xc3\xbf'
            # self.assertEqual(d(input, encoding="iso-8859-1", force=True), output)
            #
            # input = '@charset "utf-8"; \xc3\xbf'
            # output = u'@charset "utf-8"; \xff'
            # self.assertEqual(d(input, encoding="iso-8859-1", force=False), output)

            input = '@charset "utf-8"; \xff'.encode()
            output = '@charset "utf-8"; \xff'
            assert d(input) == output

            # input = b'@charset "utf-8"; \xc3\xbf'
            input = '@charset "utf-8"; \xff'.encode()
            output = '@charset "iso-8859-1"; \xc3\xbf'
            assert d(input, encoding="iso-8859-1", force=True) == output

            # input = b'\xc3\xbf'
            input = '\xff'.encode()
            output = '\xc3\xbf'
            assert d(input, encoding="iso-8859-1", force=True) == output

            # input = b'@charset "utf-8"; \xc3\xbf'
            input = '@charset "utf-8"; \xff'.encode()
            output = '@charset "utf-8"; \xff'
            assert d(input, encoding="iso-8859-1", force=False) == output



"""Testcases for cssutils.css.CSSComment"""

import xml

import cssutils.css

from . import test_cssrule


class TestCSSComment(test_cssrule.TestCSSRule):
    def _setup_rule(self):
        self.r = cssutils.css.CSSComment()
        self.rRO = cssutils.css.CSSComment(readonly=True)
        self.r_type = cssutils.css.CSSComment.COMMENT
        self.r_typeString = 'COMMENT'

    def test_csstext(self):
        "CSSComment.cssText"
        tests = {
            '/*öäüß€ÖÄÜ*/': '/*\xf6\xe4\xfc\xdf\u20ac\xd6\xc4\xdc*/',
            '/*x*/': None,
            '/* x */': None,
            '/*\t12\n*/': None,
            '/* /* */': None,
            '/* \\*/': None,
            '/*"*/': None,
            '''/*"
            */''': None,
            '/** / ** //*/': None,
        }
        self.do_equal_r(tests)  # set cssText
        tests.update({
            '/*x': '/*x*/',
            '\n /*': '/**/',
        })
        self.do_equal_p(tests)  # parse

        tests = {
            '/* */ ': xml.dom.InvalidModificationErr,
            '/* *//**/': xml.dom.InvalidModificationErr,
            '/* */1': xml.dom.InvalidModificationErr,
            '/* */ */': xml.dom.InvalidModificationErr,
            '  */ /* ': xml.dom.InvalidModificationErr,
            '*/': xml.dom.InvalidModificationErr,
            '@x /* x */': xml.dom.InvalidModificationErr,
        }
        self.do_raise_r(tests)  # set cssText
        # no raising of error possible?
        # self.do_raise_p(tests) # parse

    def test_InvalidModificationErr(self):
        "CSSComment.cssText InvalidModificationErr"
        self._test_InvalidModificationErr('/* comment */')

    def test_reprANDstr(self):
        "CSSComment.__repr__(), .__str__()"
        text = '/* test */'

        s = cssutils.css.CSSComment(cssText=text)

        s2 = eval(repr(s))
        assert isinstance(s2, s.__class__)
        assert text == s2.cssText



"""utility script to parse given filenames or string"""

import logging
import optparse
import sys

import cssutils


def main(args=None):
    """
    Parses given filename(s) or string or URL (using optional encoding) and
    prints the parsed style sheet to stdout.

    Redirect stdout to save CSS. Redirect stderr to save parser log infos.
    """
    usage = """usage: %prog [options] filename1.css [filename2.css ...]
        [>filename_combined.css] [2>parserinfo.log] """
    p = optparse.OptionParser(usage=usage)
    p.add_option(
        '-s', '--string', action='store_true', dest='string', help='parse given string'
    )
    p.add_option('-u', '--url', action='store', dest='url', help='parse given url')
    p.add_option(
        '-e',
        '--encoding',
        action='store',
        dest='encoding',
        help='encoding of the file or override encoding found',
    )
    p.add_option(
        '-m',
        '--minify',
        action='store_true',
        dest='minify',
        help='minify parsed CSS',
        default=False,
    )
    p.add_option(
        '-d',
        '--debug',
        action='store_true',
        dest='debug',
        help='activate debugging output',
    )

    (options, params) = p.parse_args(args)

    if not params and not options.url:
        p.error("no filename given")

    if options.debug:
        p = cssutils.CSSParser(loglevel=logging.DEBUG)
    else:
        p = cssutils.CSSParser()

    if options.minify:
        cssutils.ser.prefs.useMinified()

    if options.string:
        sheet = p.parseString(''.join(params), encoding=options.encoding)
        print(sheet.cssText)
    elif options.url:
        sheet = p.parseUrl(options.url, encoding=options.encoding)
        print(sheet.cssText)
    else:
        for filename in params:
            sys.stderr.write('=== CSS FILE: "%s" ===\n' % filename)
            sheet = p.parseFile(filename, encoding=options.encoding)
            print(sheet.cssText)
            print()
            sys.stderr.write('\n')


if __name__ == "__main__":
    sys.exit(main())



"""StyleSheetList implements DOM Level 2 Style Sheets StyleSheetList."""

__all__ = ['StyleSheetList']


class StyleSheetList(list):
    """Interface `StyleSheetList` (introduced in DOM Level 2)

    The `StyleSheetList` interface provides the abstraction of an ordered
    collection of :class:`~cssutils.stylesheets.StyleSheet` objects.

    The items in the `StyleSheetList` are accessible via an integral index,
    starting from 0.

    This Python implementation is based on a standard Python list so e.g.
    allows ``examplelist[index]`` usage.
    """

    def item(self, index):
        """
        Used to retrieve a style sheet by ordinal `index`. If `index` is
        greater than or equal to the number of style sheets in the list,
        this returns ``None``.
        """
        try:
            return self[index]
        except IndexError:
            return None

    length = property(
        lambda self: len(self),
        doc="The number of :class:`StyleSheet` objects in the list. The range"
        "  of valid child stylesheet indices is 0 to length-1 inclusive.",
    )



"""CSS2Properties (partly!) implements DOM Level 2 CSS CSS2Properties used
by CSSStyleDeclaration

TODO: CSS2Properties
    If an implementation does implement this interface, it is expected to
    understand the specific syntax of the shorthand properties, and apply
    their semantics; when the margin property is set, for example, the
    marginTop, marginRight, marginBottom and marginLeft properties are
    actually being set by the underlying implementation.

    When dealing with CSS "shorthand" properties, the shorthand properties
    should be decomposed into their component longhand properties as
    appropriate, and when querying for their value, the form returned
    should be the shortest form exactly equivalent to the declarations made
    in the ruleset. However, if there is no shorthand declaration that
    could be added to the ruleset without changing in any way the rules
    already declared in the ruleset (i.e., by adding longhand rules that
    were previously not declared in the ruleset), then the empty string
    should be returned for the shorthand property.

    For example, querying for the font property should not return
    "normal normal normal 14pt/normal Arial, sans-serif", when
    "14pt Arial, sans-serif" suffices. (The normals are initial values, and
    are implied by use of the longhand property.)

    If the values for all the longhand properties that compose a particular
    string are the initial values, then a string consisting of all the
    initial values should be returned (e.g. a border-width value of
    "medium" should be returned as such, not as "").

    For some shorthand properties that take missing values from other
    sides, such as the margin, padding, and border-[width|style|color]
    properties, the minimum number of sides possible should be used; i.e.,
    "0px 10px" will be returned instead of "0px 10px 0px 10px".

    If the value of a shorthand property can not be decomposed into its
    component longhand properties, as is the case for the font property
    with a value of "menu", querying for the values of the component
    longhand properties should return the empty string.

TODO: CSS2Properties DOMImplementation
    The interface found within this section are not mandatory. A DOM
    application can use the hasFeature method of the DOMImplementation
    interface to determine whether it is supported or not. The feature
    string for this extended interface listed in this section is "CSS2"
    and the version is "2.0".

"""

__all__ = ['CSS2Properties']

import re

import cssutils.profiles


class CSS2Properties:
    """The CSS2Properties interface represents a convenience mechanism
    for retrieving and setting properties within a CSSStyleDeclaration.
    The attributes of this interface correspond to all the properties
    specified in CSS2. Getting an attribute of this interface is
    equivalent to calling the getPropertyValue method of the
    CSSStyleDeclaration interface. Setting an attribute of this
    interface is equivalent to calling the setProperty method of the
    CSSStyleDeclaration interface.

    cssutils actually also allows usage of ``del`` to remove a CSS property
    from a CSSStyleDeclaration.

    This is an abstract class, the following functions need to be present
    in inheriting class:

    - ``_getP``
    - ``_setP``
    - ``_delP``
    """

    # actual properties are set after the class definition!
    def _getP(self, CSSname):
        pass

    def _setP(self, CSSname, value):
        pass

    def _delP(self, CSSname):
        pass


_reCSStoDOMname = re.compile('-[a-z]', re.I)


def _toDOMname(CSSname):
    """Returns DOMname for given CSSname e.g. for CSSname 'font-style' returns
    'fontStyle'.
    """

    def _doCSStoDOMname2(m):
        return m.group(0)[1].capitalize()

    return _reCSStoDOMname.sub(_doCSStoDOMname2, CSSname)


_reDOMtoCSSname = re.compile('([A-Z])[a-z]+')


def _toCSSname(DOMname):
    """Return CSSname for given DOMname e.g. for DOMname 'fontStyle' returns
    'font-style'.
    """

    def _doDOMtoCSSname2(m):
        return '-' + m.group(0).lower()

    return _reDOMtoCSSname.sub(_doDOMtoCSSname2, DOMname)


# add list of DOMname properties to CSS2Properties
# used for CSSStyleDeclaration to check if allowed properties
# but somehow doubled, any better way?
CSS2Properties._properties = []
for group in cssutils.profiles.properties:
    for name in cssutils.profiles.properties[group]:
        CSS2Properties._properties.append(_toDOMname(name))


# add CSS2Properties to CSSStyleDeclaration:
def __named_property_def(DOMname):
    """
    Closure to keep name known in each properties accessor function
    DOMname is converted to CSSname here, so actual calls use CSSname.
    """
    CSSname = _toCSSname(DOMname)

    def _get(self):
        return self._getP(CSSname)

    def _set(self, value):
        self._setP(CSSname, value)

    def _del(self):
        self._delP(CSSname)

    return _get, _set, _del


# add all CSS2Properties to CSSStyleDeclaration
for DOMname in CSS2Properties._properties:
    setattr(CSS2Properties, DOMname, property(*__named_property_def(DOMname)))



"""Testcases for cssutils.css.CSSImportRule"""

import re
import xml.dom

import pytest

import cssutils

from . import test_cssrule


class TestCSSImportRule(test_cssrule.TestCSSRule):
    def _setup_rule(self):
        self.r = cssutils.css.CSSImportRule()
        self.rRO = cssutils.css.CSSImportRule(readonly=True)
        self.r_type = cssutils.css.CSSImportRule.IMPORT_RULE
        self.r_typeString = 'IMPORT_RULE'

    def test_init(self):
        "CSSImportRule.__init__()"
        super().test_init()

        # no init param
        assert self.r.href is None
        assert self.r.hreftype is None
        assert not self.r.hrefFound
        assert 'all' == self.r.media.mediaText
        assert isinstance(self.r.media, cssutils.stylesheets.MediaList)
        assert self.r.name is None
        assert isinstance(self.r.styleSheet, cssutils.css.CSSStyleSheet)
        assert 0 == self.r.styleSheet.cssRules.length
        assert '' == self.r.cssText

        # all
        r = cssutils.css.CSSImportRule(href='href', mediaText='tv', name='name')
        assert '@import url(href) tv "name";' == r.cssText
        assert "href" == r.href
        assert r.hreftype is None
        assert 'tv' == r.media.mediaText
        assert isinstance(r.media, cssutils.stylesheets.MediaList)
        assert 'name' == r.name
        assert r.parentRule is None  # see CSSRule
        assert r.parentStyleSheet is None  # see CSSRule
        assert isinstance(self.r.styleSheet, cssutils.css.CSSStyleSheet)
        assert 0 == self.r.styleSheet.cssRules.length

        # href
        r = cssutils.css.CSSImportRule('x')
        assert '@import url(x);' == r.cssText
        assert 'x' == r.href
        assert r.hreftype is None

        # href + mediaText
        r = cssutils.css.CSSImportRule('x', 'print')
        assert '@import url(x) print;' == r.cssText
        assert 'x' == r.href
        assert 'print' == r.media.mediaText

        # href + name
        r = cssutils.css.CSSImportRule('x', name='n')
        assert '@import url(x) "n";' == r.cssText
        assert 'x' == r.href
        assert 'n' == r.name

        # href + mediaText + name
        r = cssutils.css.CSSImportRule('x', 'print', 'n')
        assert '@import url(x) print "n";' == r.cssText
        assert 'x' == r.href
        assert 'print' == r.media.mediaText
        assert 'n' == r.name

        # media +name only
        self.r = cssutils.css.CSSImportRule(mediaText='print', name="n")
        assert isinstance(self.r.media, cssutils.stylesheets.MediaList)
        assert '' == self.r.cssText
        assert 'print' == self.r.media.mediaText
        assert 'n' == self.r.name

        # only possible to set @... similar name
        with pytest.raises(xml.dom.InvalidModificationErr):
            self.r._setAtkeyword('x')

    def test_cssText(self):
        "CSSImportRule.cssText"
        tests = {
            # href string
            '''@import "str";''': None,
            '''@import"str";''': '''@import "str";''',
            '''@\\import "str";''': '''@import "str";''',
            '''@IMPORT "str";''': '''@import "str";''',
            '''@import 'str';''': '''@import "str";''',
            '''@import 'str' ;''': '''@import "str";''',
            '''@import "str"  ;''': '''@import "str";''',
            r'''@import "\""  ;''': r'''@import "\"";''',
            '''@import '\\'';''': r'''@import "'";''',
            '''@import '"';''': r'''@import "\"";''',
            # href url
            '''@import url(x.css);''': None,
            # nospace
            '''@import url(")");''': '''@import url(")");''',
            '''@import url("\\"");''': '''@import url("\\"");''',
            '''@import url('\\'');''': '''@import url("'");''',
            # href + media
            # all is removed
            '''@import "str" all;''': '''@import "str";''',
            '''@import "str" tv, print;''': None,
            '''@import"str"tv,print;''': '''@import "str" tv, print;''',
            '''@import "str" tv, print, all;''': '''@import "str";''',
            '''@import "str" handheld, all;''': '''@import "str";''',
            '''@import "str" all, handheld;''': '''@import "str";''',
            '''@import "str" not tv;''': None,
            '''@import "str" only tv;''': None,
            '''@import "str" only tv and (color: 2);''': None,
            # href + name
            '''@import "str" "name";''': None,
            '''@import "str" 'name';''': '''@import "str" "name";''',
            '''@import url(x) "name";''': None,
            '''@import "str" "\\"";''': None,
            '''@import "str" '\\'';''': '''@import "str" "'";''',
            # href + media + name
            '''@import"str"tv"name";''': '''@import "str" tv "name";''',
            '''@import\t\r\f\n"str"\t\t\r\f\ntv\t\t\r\f\n"name"\t;''': '''@import "str" tv "name";''',
            # comments
            '''@import /*1*/ "str" /*2*/;''': None,
            '@import/*1*//*2*/"str"/*3*//*4*/all/*5*//*6*/"name"/*7*//*8*/ ;': '@import /*1*/ /*2*/ "str" /*3*/ /*4*/ all /*5*/ /*6*/ "name" /*7*/ /*8*/;',
            '@import/*1*//*2*/url(u)/*3*//*4*/all/*5*//*6*/"name"/*7*//*8*/ ;': '@import /*1*/ /*2*/ url(u) /*3*/ /*4*/ all /*5*/ /*6*/ "name" /*7*/ /*8*/;',
            '@import/*1*//*2*/url("u")/*3*//*4*/all/*5*//*6*/"name"/*7*//*8*/ ;': '@import /*1*/ /*2*/ url(u) /*3*/ /*4*/ all /*5*/ /*6*/ "name" /*7*/ /*8*/;',
            # WS
            '@import\n\t\f "str"\n\t\f tv\n\t\f "name"\n\t\f ;': '@import "str" tv "name";',
            '@import\n\t\f url(\n\t\f u\n\t\f )\n\t\f tv\n\t\f "name"\n\t\f ;': '@import url(u) tv "name";',
            '@import\n\t\f url("u")\n\t\f tv\n\t\f "name"\n\t\f ;': '@import url(u) tv "name";',
            '@import\n\t\f url(\n\t\f "u"\n\t\f )\n\t\f tv\n\t\f "name"\n\t\f ;': '@import url(u) tv "name";',
        }
        self.do_equal_r(tests)  # set cssText
        tests.update({
            '@import "x.css" tv': '@import "x.css" tv;',
            '@import "x.css"': '@import "x.css";',  # no ;
            "@import 'x.css'": '@import "x.css";',  # no ;
            '@import url(x.css)': '@import url(x.css);',  # no ;
            '@import "x;': '@import "x;";',  # no "!
        })
        self.do_equal_p(tests)  # parse

        tests = {
            '''@import;''': xml.dom.SyntaxErr,
            '''@import all;''': xml.dom.SyntaxErr,
            '''@import all"name";''': xml.dom.SyntaxErr,
            '''@import x";''': xml.dom.SyntaxErr,
            '''@import "str" ,all;''': xml.dom.SyntaxErr,
            '''@import "str" all,;''': xml.dom.SyntaxErr,
            '''@import "str" all tv;''': xml.dom.SyntaxErr,
            '''@import "str" "name" all;''': xml.dom.SyntaxErr,
        }
        self.do_raise_p(tests)  # parse
        tests.update({
            '@import "x.css"': xml.dom.SyntaxErr,
            "@import 'x.css'": xml.dom.SyntaxErr,
            '@import url(x.css)': xml.dom.SyntaxErr,
            '@import "x.css" tv': xml.dom.SyntaxErr,
            '@import "x;': xml.dom.SyntaxErr,
            '''@import url("x);''': xml.dom.SyntaxErr,
            # trailing
            '''@import "x";"a"''': xml.dom.SyntaxErr,
            # trailing S or COMMENT
            '''@import "x";/**/''': xml.dom.SyntaxErr,
            '''@import "x"; ''': xml.dom.SyntaxErr,
        })
        self.do_raise_r(tests)  # set cssText

    def test_href(self):
        "CSSImportRule.href"
        # set
        self.r.href = 'x'
        assert 'x' == self.r.href
        assert '@import url(x);' == self.r.cssText

        # http
        self.r.href = 'http://www.example.com/x?css=z&v=1'
        assert 'http://www.example.com/x?css=z&v=1' == self.r.href
        assert '@import url(http://www.example.com/x?css=z&v=1);' == self.r.cssText

        # also if hreftype changed
        self.r.hreftype = 'string'
        assert 'http://www.example.com/x?css=z&v=1' == self.r.href
        assert '@import "http://www.example.com/x?css=z&v=1";' == self.r.cssText

        # string escaping?
        self.r.href = '"'
        assert '@import "\\"";' == self.r.cssText
        self.r.hreftype = 'url'
        assert '@import url("\\"");' == self.r.cssText

        # url escaping?
        self.r.href = ')'
        assert '@import url(")");' == self.r.cssText

        self.r.hreftype = 'NOT VALID'  # using default
        assert '@import url(")");' == self.r.cssText

    def test_hrefFound(self):
        "CSSImportRule.hrefFound"

        def fetcher(url):
            if url == 'http://example.com/yes':
                return None, '/**/'
            else:
                return None, None

        parser = cssutils.CSSParser(fetcher=fetcher)
        sheet = parser.parseString('@import "http://example.com/yes" "name"')

        r = sheet.cssRules[0]
        assert b'/**/' == r.styleSheet.cssText
        assert r.hrefFound
        assert 'name' == r.name

        r.cssText = '@import url(http://example.com/none) "name2";'
        assert b'' == r.styleSheet.cssText
        assert not r.hrefFound
        assert 'name2' == r.name

        sheet.cssText = '@import url(http://example.com/none);'
        assert r != sheet.cssRules[0]

    def test_hreftype(self):
        "CSSImportRule.hreftype"
        self.r = cssutils.css.CSSImportRule()

        self.r.cssText = '@import /*1*/url(org) /*2*/;'
        assert 'uri' == self.r.hreftype
        assert '@import /*1*/ url(org) /*2*/;' == self.r.cssText

        self.r.cssText = '@import /*1*/"org" /*2*/;'
        assert 'string' == self.r.hreftype
        assert '@import /*1*/ "org" /*2*/;' == self.r.cssText

        self.r.href = 'new'
        assert '@import /*1*/ "new" /*2*/;' == self.r.cssText

        self.r.hreftype = 'uri'
        assert '@import /*1*/ url(new) /*2*/;' == self.r.cssText

    def test_media(self):
        "CSSImportRule.media"
        self.r.href = 'x'  # @import url(x)

        # media is readonly
        with pytest.raises(AttributeError):
            self.r.__setattr__('media', None)

        # but not static
        self.r.media.mediaText = 'print'
        assert '@import url(x) print;' == self.r.cssText
        self.r.media.appendMedium('tv')
        assert '@import url(x) print, tv;' == self.r.cssText

        tv_msg = re.escape(
            '''MediaList: Ignoring new medium '''
            '''cssutils.stylesheets.MediaQuery(mediaText='tv') '''
            '''as already specified "all" (set ``mediaText`` instead).'''
        )

        # for generated rule
        r = cssutils.css.CSSImportRule(href='x')
        with pytest.raises(xml.dom.InvalidModificationErr, match=tv_msg):
            r.media.appendMedium('tv')
        assert '@import url(x);' == r.cssText
        with pytest.raises(xml.dom.InvalidModificationErr, match=tv_msg):
            r.media.appendMedium('tv')
        assert '@import url(x);' == r.cssText
        r.media.mediaText = 'tv'
        assert '@import url(x) tv;' == r.cssText
        r.media.appendMedium('print')  # all + tv = all!
        assert '@import url(x) tv, print;' == r.cssText

        # for parsed rule without initial media
        s = cssutils.parseString('@import url(x);')
        r = s.cssRules[0]

        with pytest.raises(xml.dom.InvalidModificationErr, match=tv_msg):
            r.media.appendMedium('tv')
        assert '@import url(x);' == r.cssText
        with pytest.raises(xml.dom.InvalidModificationErr, match=tv_msg):
            r.media.appendMedium('tv')
        assert '@import url(x);' == r.cssText
        r.media.mediaText = 'tv'
        assert '@import url(x) tv;' == r.cssText
        r.media.appendMedium('print')  # all + tv = all!
        assert '@import url(x) tv, print;' == r.cssText

    def test_name(self):
        "CSSImportRule.name"
        r = cssutils.css.CSSImportRule('x', name='a000000')
        assert 'a000000' == r.name
        assert '@import url(x) "a000000";' == r.cssText

        r.name = "n"
        assert 'n' == r.name
        assert '@import url(x) "n";' == r.cssText
        r.name = '"'
        assert '"' == r.name
        assert '@import url(x) "\\"";' == r.cssText

        r.hreftype = 'string'
        assert '@import "x" "\\"";' == r.cssText
        r.name = "123"
        assert '@import "x" "123";' == r.cssText

        r.name = None
        assert r.name is None
        assert '@import "x";' == r.cssText

        r.name = ""
        assert r.name is None
        assert '@import "x";' == r.cssText

        with pytest.raises(xml.dom.SyntaxErr):
            r._setName(0)
        with pytest.raises(xml.dom.SyntaxErr):
            r._setName(123)

    def test_styleSheet(self):
        "CSSImportRule.styleSheet"

        def fetcher(url):
            if url == "/root/level1/anything.css":
                return None, '@import "level2/css.css" "title2";'
            else:
                return None, 'a { color: red }'

        parser = cssutils.CSSParser(fetcher=fetcher)
        sheet = parser.parseString(
            '''@charset "ascii";
                                   @import "level1/anything.css" tv "title";''',
            href='/root/',
        )

        assert sheet.href == '/root/'

        ir = sheet.cssRules[1]
        assert ir.href == 'level1/anything.css'
        assert ir.styleSheet.href == '/root/level1/anything.css'
        # inherits ascii as no self charset is set
        assert ir.styleSheet.encoding == 'ascii'
        assert ir.styleSheet.ownerRule == ir
        assert ir.styleSheet.media.mediaText == 'tv'
        assert ir.styleSheet.parentStyleSheet is None  # sheet
        assert ir.styleSheet.title == 'title'
        assert (
            ir.styleSheet.cssText
            == b'@charset "ascii";\n@import "level2/css.css" "title2";'
        )

        ir2 = ir.styleSheet.cssRules[1]
        assert ir2.href == 'level2/css.css'
        assert ir2.styleSheet.href == '/root/level1/level2/css.css'
        # inherits ascii as no self charset is set
        assert ir2.styleSheet.encoding == 'ascii'
        assert ir2.styleSheet.ownerRule == ir2
        assert ir2.styleSheet.media.mediaText == 'all'
        assert ir2.styleSheet.parentStyleSheet is None  # ir.styleSheet
        assert ir2.styleSheet.title == 'title2'
        assert (
            ir2.styleSheet.cssText == b'@charset "ascii";\na {\n    color: red\n    }'
        )

        sheet = cssutils.parseString('@import "CANNOT-FIND.css";')
        ir = sheet.cssRules[0]
        assert ir.href == "CANNOT-FIND.css"
        assert isinstance(ir.styleSheet, cssutils.css.CSSStyleSheet)

        def fetcher(url):
            if url.endswith('level1.css'):
                return None, b'@charset "ascii"; @import "level2.css";'
            else:
                return None, b'a { color: red }'

        parser = cssutils.CSSParser(fetcher=fetcher)

        sheet = parser.parseString('@charset "iso-8859-1";@import "level1.css";')
        assert sheet.encoding == 'iso-8859-1'

        sheet = sheet.cssRules[1].styleSheet
        assert sheet.encoding == 'ascii'

        sheet = sheet.cssRules[1].styleSheet
        assert sheet.encoding == 'ascii'

    def test_incomplete(self):
        "CSSImportRule (incomplete)"
        tests = {
            '@import "x.css': '@import "x.css";',
            "@import 'x": '@import "x";',
            # TODO:
            "@import url(x": '@import url(x);',
            "@import url('x": '@import url(x);',
            '@import url("x;': '@import url("x;");',
            '@import url( "x;': '@import url("x;");',
            '@import url("x ': '@import url("x ");',
            '@import url(x ': '@import url(x);',
            '''@import "a
                @import "b";
                @import "c";''': '@import "c";',
        }
        self.do_equal_p(tests, raising=False)  # parse

    def test_InvalidModificationErr(self):
        "CSSImportRule.cssText InvalidModificationErr"
        self._test_InvalidModificationErr('@import')

    def test_reprANDstr(self):
        "CSSImportRule.__repr__(), .__str__()"
        href = 'x.css'
        mediaText = 'tv, print'
        name = 'name'
        s = cssutils.css.CSSImportRule(href=href, mediaText=mediaText, name=name)

        # str(): mediaText nor name are present here
        assert href in str(s)

        # repr()
        s2 = eval(repr(s))
        assert isinstance(s2, s.__class__)
        assert href == s2.href
        assert mediaText == s2.media.mediaText
        assert name == s2.name



"""Testcases for cssutils.css.CSSValue and CSSPrimitiveValue."""

import re

import pytest

import cssutils

CSS2 = (cssutils.profile.CSS_LEVEL_2,)
C3BUI = (cssutils.profile.CSS3_BASIC_USER_INTERFACE,)
C3BB = (cssutils.profile.CSS3_BACKGROUNDS_AND_BORDERS,)
CM3 = (cssutils.profile.CSS3_COLOR,)
FM3 = (cssutils.profile.CSS3_FONTS,)
C3PM = (cssutils.profile.CSS3_PAGED_MEDIA,)
C3T = (cssutils.profile.CSS3_TEXT,)
FM3FF = (cssutils.profile.CSS3_FONT_FACE,)
CSS2_CM3 = (CM3[0], CSS2[0])
CSS2_FM3 = (FM3[0], CSS2[0])


class TestProfiles:
    M1 = {'testvalue': 'x'}
    P1 = {
        '-test-tokenmacro': '({num}{w}){1,2}',
        '-test-macro': '{ident}|{percentage}',
        '-test-custommacro': '{testvalue}',
        # custom validation function
        '-test-funcval': lambda v: int(v) > 0,
    }

    def test_knownNames(self):
        "Profiles.knownNames"
        p = cssutils.profiles.Profiles()
        p.removeProfile(all=True)
        p.addProfile('test', self.P1, self.M1)
        assert p.knownNames == list(self.P1.keys())
        p.removeProfile(all=True)
        assert p.knownNames == []

    def test_profiles(self):
        "Profiles.profiles"
        p = cssutils.profiles.Profiles()
        p.removeProfile(all=True)
        p.addProfile('test', self.P1, self.M1)
        assert p.profiles == ['test']
        p.removeProfile(all=True)
        assert p.profiles == []

    def test_validate2(self):
        "Profiles.validate()"
        # save
        saved = cssutils.profile

        # test
        p = cssutils.profiles.Profiles()
        cssutils.profile = p

        pvs = [('color', 'red'), ('color', 'rgba(0,0,0,0)'), ('color', 'XXX')]

        def check(*results):
            for i, pv in enumerate(pvs):
                assert p.validate(*pv) == results[i]

        check(True, True, False)

        p.removeProfile(p.CSS3_COLOR)
        check(True, False, False)

        cssutils.profile.addProfile('test', {}, {'color': 'XXX'})
        check(False, False, True)

        p.removeProfile(all=True)
        check(False, False, False)

        # TODO: validateWithProfile

        # restore
        cssutils.profile = saved

    @pytest.mark.usefixtures('saved_profiles')
    def test_addProfile(self):
        "Profiles.addProfile with custom validation function"
        # unknown profile
        with pytest.raises(cssutils.profiles.NoSuchProfileException):
            list(cssutils.profile.propertiesByProfile('NOTSET'))

        # new profile
        cssutils.profile.addProfile('test', self.P1, self.M1)

        props = list(self.P1.keys())
        props.sort()
        assert props == list(cssutils.profile.propertiesByProfile('test'))

        cssutils.log.raiseExceptions = False
        tests = {
            ('-test-tokenmacro', '1'): True,
            ('-test-tokenmacro', '1 -2'): True,
            ('-test-tokenmacro', '1 2 3'): False,
            ('-test-tokenmacro', 'a'): False,
            ('-test-macro', 'a'): True,
            ('-test-macro', '0.1%'): True,
            ('-test-custommacro', 'x'): True,
            ('-test-custommacro', '1'): False,
            ('-test-custommacro', 'y'): False,
            ('-test-funcval', '1'): True,
            ('-test-funcval', '-1'): False,
            ('-test-funcval', 'x'): False,
        }
        for test, v in list(tests.items()):
            assert v == cssutils.profile.validate(*test)

            assert (v, v, ['test']) == cssutils.profile.validateWithProfile(*test)

        cssutils.log.raiseExceptions = True

        expmsg = re.escape("invalid literal for int() with base 10: 'x'")
        with pytest.raises(Exception, match=expmsg):
            cssutils.profile.validate('-test-funcval', 'x')

    def test_removeProfile(self):
        "Profiles.removeProfile()"
        p = cssutils.profiles.Profiles()
        assert 9 == len(p.profiles)
        p.removeProfile(p.CSS_LEVEL_2)
        assert 8 == len(p.profiles)
        p.removeProfile(all=True)
        assert 0 == len(p.profiles)

    def _gen_validation_inputs():
        CSS_LEVEL_2 = cssutils.profiles.Profiles.CSS_LEVEL_2
        CSS3_COLOR = cssutils.profiles.Profiles.CSS3_COLOR
        yield (
            ('color', 'red', None),
            (True, True, [CSS_LEVEL_2]),
        )
        yield (
            ('color', 'red', CSS_LEVEL_2),
            (True, True, [CSS_LEVEL_2]),
        )
        yield (
            ('color', 'red', CSS3_COLOR),
            (True, False, [CSS_LEVEL_2]),
        )
        yield pytest.param(
            ('color', 'rgba(0,0,0,0)', None),
            (True, True, [CSS3_COLOR]),
            marks=pytest.mark.xfail(reason="#37"),
        )
        yield pytest.param(
            ('color', 'rgba(0,0,0,0)', CSS_LEVEL_2),
            (True, False, [CSS3_COLOR]),
            marks=pytest.mark.xfail(reason="#37"),
        )
        yield pytest.param(
            ('color', 'rgba(0,0,0,0)', CSS3_COLOR),
            (True, True, [CSS3_COLOR]),
            marks=pytest.mark.xfail(reason="#37"),
        )
        yield pytest.param(
            ('color', '1px', None),
            (
                False,
                False,
                [CSS3_COLOR, CSS_LEVEL_2],
            ),
            marks=pytest.mark.xfail(reason="#37"),
        )
        yield pytest.param(
            ('color', '1px', CSS_LEVEL_2),
            (
                False,
                False,
                [CSS3_COLOR, CSS_LEVEL_2],
            ),
            marks=pytest.mark.xfail(reason="#37"),
        )
        yield pytest.param(
            ('color', '1px', CSS3_COLOR),
            (
                False,
                False,
                [CSS3_COLOR, CSS_LEVEL_2],
            ),
            marks=pytest.mark.xfail(reason="#37"),
        )
        yield (
            ('color', 'aliceblue', None),
            (True, True, [CSS_LEVEL_2]),
        )
        yield (
            ('opacity', '1', None),
            (True, True, [CSS3_COLOR]),
        )
        yield (
            ('opacity', '1', CSS_LEVEL_2),
            (True, False, [CSS3_COLOR]),
        )
        yield (
            ('opacity', '1', CSS3_COLOR),
            (True, True, [CSS3_COLOR]),
        )
        yield (
            ('opacity', '1px', None),
            (False, False, [CSS3_COLOR]),
        )
        yield (
            ('opacity', '1px', CSS_LEVEL_2),
            (False, False, [CSS3_COLOR]),
        )
        yield (
            ('opacity', '1px', CSS3_COLOR),
            (False, False, [CSS3_COLOR]),
        )
        yield (
            ('-x', '1', None),
            (False, False, []),
        )
        yield (
            ('-x', '1', CSS_LEVEL_2),
            (False, False, []),
        )
        yield (
            ('-x', '1', CSS3_COLOR),
            (False, False, []),
        )

    @pytest.mark.parametrize(('params', 'results'), _gen_validation_inputs())
    def test_validateWithProfile(self, params, results):
        "Profiles.validate(), Profiles.validateWithProfile()"
        p = cssutils.profiles.Profiles()
        assert p.validate(*params[:2]) == results[0]
        assert p.validateWithProfile(*params) == results

    def test_propertiesByProfile(self):
        "Profiles.propertiesByProfile"
        assert ['opacity'] == list(
            cssutils.profile.propertiesByProfile(cssutils.profile.CSS3_COLOR)
        )

    def test_csscolorlevel3(self):
        "CSS Color Module Level 3"
        # (propname, propvalue): (valid, validprofile)
        namedcolors = '''transparent, orange,
                         aqua, black, blue, fuchsia, gray, green, lime, maroon,
                         navy, olive, purple, red, silver, teal, white, yellow'''
        for color in namedcolors.split(','):
            color = color.strip()
            assert cssutils.profile.validate('color', color)

            assert (True, True, list(CSS2)) == cssutils.profile.validateWithProfile(
                'color', color
            )

        # CSS2 only:
        uicolor = (
            'ActiveBorder|ActiveCaption|AppWorkspace|Background|ButtonFace|'
            'ButtonHighlight|ButtonShadow|ButtonText|CaptionText|GrayText|Highlight|'
            'HighlightText|InactiveBorder|InactiveCaption|InactiveCaptionText|'
            'InfoBackground|InfoText|Menu|MenuText|Scrollbar|ThreeDDarkShadow|'
            'ThreeDFace|ThreeDHighlight|ThreeDLightShadow|ThreeDShadow|Window|'
            'WindowFrame|WindowText'
        )
        for color in uicolor.split('|'):
            assert not cssutils.profile.validate('color', color)

            # TODO: fixme
            # assert (True, True, list(CSS2)) == cssutils.profile.validateWithProfile(
            #     'color', color
            # )

    def test_validate(self):
        "Profiles.validate()"
        tests = {
            # name, values: valid, matching, profile
            # background-position
            (
                'background-position',
                (
                    'inherit',
                    '0',
                    '1%',
                    '1px',
                    '0 0',
                    '1% 1%',
                    '1px 1px',
                    '1px 1%',
                    'top',
                    'bottom',
                    'left',
                    'right',
                    'center center',
                    'center',
                    'top left',
                    'top center',
                    'top right',
                    'bottom left',
                    'bottom center',
                    'bottom right',
                    'center left',
                    'center center',
                    'center right',
                    '0 center',
                    'center 0',
                    '0 top',
                    '10% bottom',
                    'left 0',
                    'right 10%',
                    '1% center',
                    'center 1%',
                ),
            ): (True, True, CSS2),
            ('background-position', ('0 left', 'top 0')): (False, False, CSS2),
            (
                'border-top-right-radius',
                (
                    '1px',
                    '1%',
                    '1% -1px',
                    '1% 0',
                ),
            ): (True, True, C3BB),
            ('border-top-right-radius', ('1px 2px 2px', '/ 1px', 'black')): (
                False,
                False,
                C3BB,
            ),
            (
                'border-radius',
                (
                    '1px',
                    '1%',
                    '0',
                    '1px 1px',
                    '1px/ 1px',
                    '1px /1px',
                    '1px  /  1px',
                    '1px 1px 1px 1px',
                    '1px 1px 1px 1px / 1px 1px 1px 1px',
                ),
            ): (True, True, C3BB),
            (
                'border-radius',
                (
                    '1px /',
                    '/ 1px',
                    '1px / 1px / 1px',
                    '1px 1px 1px 1px 1px',
                    '1px / 1px 1px 1px 1px 1px',
                    'black',
                ),
            ): (False, False, C3BB),
            (
                'border',
                (
                    '1px',
                    'solid',
                    'red',
                    '1px solid red',
                    '1px red solid',
                    'red 1px solid',
                    'red solid 1px',
                    'solid 1px red',
                    'solid red 1px',
                ),
            ): (True, True, C3BB),
            (
                'border',
                (
                    '1px 1px',
                    'red red 1px',
                ),
            ): (False, False, C3BB),
            (
                'box-shadow',
                (
                    'none',
                    '1px 1px',
                    '1px 1px 1px',
                    '1px 1px 1px 1px',
                    '1px 1px 1px 1px red',
                    'inset 1px 1px',
                    'inset 1px 1px 1px 1px black',
                ),
            ): (True, True, C3BB),
            (
                'box-shadow',
                (
                    '1px',
                    '1px 1px 1px 1px 1px',
                    'x 1px 1px',
                    'inset',
                    '1px black',
                    'black',
                ),
            ): (False, False, C3BB),
            # color
            (
                'color',
                (
                    'x',
                    '#',
                    '#0',
                    '#00',
                    '#0000',
                    '#00000',
                    '#0000000',
                    '#00j',
                    '#j00000',
                    'rgb(0.0,1,1)',
                    'rgb(0)',
                    'rgb(0, 1)',
                    'rgb(0, 1, 1, 1)',
                    'rgb(0, 1, 0%)',
                    'rgba(0)',
                    'rgba(0, 1, 1.0, 1)',
                    'rgba(0, 1)',
                    'rgba(0, 1, 1, 1, 1)',
                    'rgba(100%, 0%, 0%, 1%)',
                    'rgba(100%, 0%, 0, 1)',
                    'hsl(1.5,1%,1%)',
                    'hsl(1,1,1%)',
                    'hsl(1,1%,1)',
                    'hsla(1.5,1%,1%, 1)',
                    'hsla(1,1,1%, 1)',
                    'hsla(1,1%,1, 1)',
                    'hsla(1,1%,1%, 1%)',
                ),
            ): (False, False, CSS2_CM3),
            (
                'color',
                (
                    'inherit',
                    'black',
                    '#000',
                    '#000000',
                    'rgb(0,1,1)',
                    'rgb( 0 , 1 , 1 )',
                    'rgb(-10,555,1)',
                    'rgb(100%, 1.5%, 0%)',
                    'rgb(150%, -20%, 0%)',
                ),
            ): (True, True, CSS2),
            (
                'color',
                (
                    'currentcolor',
                    'aliceblue',
                    'rgba(1,1,1,1)',
                    'rgba( 1 , 1 , 1 , 1 )',
                    'rgba(100%, 0%, 0%, 1)',
                    'hsl(1,1%,1%)',
                    'hsl( 1 , 1% , 1% )',
                    'hsl(-1000,555.5%,-61.5%)',
                    'hsla(1,1%,1%,1)',
                    'hsla( 1, 1% , 1% , 1 )',
                    'hsla(-1000,555.5%,-61.5%, 0.5)',
                ),
            ): (True, True, CM3),
            (
                'content',
                (
                    'none',
                    'normal',
                    '""',
                    "'x'",
                ),
            ): (True, True, CSS2),
            (
                'cursor',
                ('url(1), auto', 'url(1) 2 3, help', 'wait', 'inherit', 'none'),
            ): (True, True, C3BUI),
            ('cursor', ('url(1), auto, wait', 'url(1) 2, help', '1')): (
                False,
                False,
                C3BUI,
            ),
            # FONTS
            ('font-family', ('serif, x',)): (True, True, CSS2),  # CSS2_FM3),
            (
                'font-family',
                (
                    'inherit',
                    'a, b',
                    'a,b,c',
                    'a, "b", c',
                    '"a", b, "c"',
                    '"a", "b", "c"',
                    '"x y"',
                    'serif',
                    # valid but CSS2: font with name serif, CSS3: same as `serif`
                    '"serif"',
                    'a  b',  # should use quotes but valid
                    'a, b   b, d',
                ),
            ): (True, True, CSS2),
            (
                'font-weight',
                (
                    'normal',
                    'bold',
                    'bolder',
                    'lighter',
                    'inherit',
                    '100',
                    '200',
                    '300',
                    '400',
                    '500',
                    '600',
                    '700',
                    '800',
                    '900',
                ),
            ): (True, True, CSS2),
            (
                'font-stretch',
                (
                    'normal',
                    'wider',
                    'narrower',
                    'ultra-condensed',
                    'extra-condensed',
                    'condensed',
                    'semi-condensed',
                    'semi-expanded',
                    'expanded',
                    'extra-expanded',
                    'ultra-expanded',
                    'inherit',
                ),
            ): (True, True, FM3),
            ('font-style', ('normal', 'italic', 'oblique', 'inherit')): (
                True,
                True,
                CSS2,
            ),
            ('font-variant', ('normal', 'small-caps', 'inherit')): (True, True, CSS2),
            ('font-size', ('-1em',)): (False, False, CSS2),
            (
                'font-size',
                (
                    'xx-small',
                    'x-small',
                    'small',
                    'medium',
                    'large',
                    'x-large',
                    'xx-large',
                    'larger',
                    'smaller',
                    '1em',
                    '1%',
                    'inherit',
                ),
            ): (True, True, CSS2),
            ('font-size-adjust', ('1.0', 'none', 'inherit')): (True, True, FM3),
            (
                'font',
                (
                    'italic small-caps bold 1px/3 a, "b", serif',
                    '12pt/14pt sans-serif',
                    '80% sans-serif',
                    'x-large/110% "new century schoolbook", serif',
                    'bold italic large Palatino, serif',
                    'normal small-caps 120%/120% fantasy',
                    'oblique 12pt "Helvetica Nue", serif',
                    'caption',
                    'icon',
                    'menu',
                    'message-box',
                    'small-caption',
                    'status-bar',
                    'inherit',
                ),
            ): (True, True, CSS2),
            ('nav-index', ('1', 'auto', 'inherit')): (True, True, C3BUI),
            ('nav-index', ('x', '1 2', '1px')): (False, False, C3BUI),
            (
                'opacity',
                (
                    'inherit',
                    '0',
                    '0.0',
                    '0.42342',
                    '1',
                    '1.0',
                    # should be clipped but valid
                    '-0',
                    '-0.1',
                    '-10',
                    '2',
                ),
            ): (True, True, CM3),
            ('opacity', ('a', '#000', '+1')): (False, False, CM3),
            (
                'outline',
                (
                    'red dotted 1px',
                    'dotted 1px red',
                    '1px red dotted',
                    'red',
                    '1px',
                    'dotted',
                    'red 1px',
                    '1px dotted',
                    'red dotted',
                    'inherit',
                ),
            ): (True, True, C3BUI),
            ('outline', ('red #fff', 'solid dotted', 'Url(x)', '1px 1px')): (
                False,
                False,
                C3BUI,
            ),
            ('outline-color', ('red', '#fff', 'inherit')): (True, True, C3BUI),
            ('outline-color', ('0', '1em')): (False, False, C3BUI),
            ('outline-offset', ('0', '1em', 'inherit')): (True, True, C3BUI),
            ('outline-offset', ('1%', 'red')): (False, False, C3BUI),
            ('outline-style', ('auto', 'dotted', 'inherit')): (True, True, C3BUI),
            ('outline-style', ('0', '1em', 'red')): (False, False, C3BUI),
            ('outline-width', ('0', '1em', 'inherit')): (True, True, C3BUI),
            ('outline-width', ('auto', 'red', 'dotted')): (False, False, C3BUI),
            ('resize', ('none', 'both', 'horizontal', 'vertical', 'inherit')): (
                True,
                True,
                C3BUI,
            ),
            ('resize', ('1', 'auto', '1px', '2%')): (False, False, C3BUI),
            (
                'size',
                (
                    '1cm',
                    '1mm 20cm',
                    'auto',
                    'landscape letter',
                    'a4 portrait',
                    'landscape',
                    'a5',
                    # 'inherit'
                ),
            ): (True, True, C3PM),
            ('size', ('portrait landscape', 'a5 letter', '2%')): (False, False, C3PM),
            (
                'src',
                (
                    'url(  a  )',
                    'local(  x  )',
                    'local("x")',
                    'local(  "x"  )',
                    'url(../fonts/LateefRegAAT.ttf) format(  "truetype-aat"  )',
                    'url(a) format(  "123x"  , "a"   )',
                    'url(a) format( "123x"  , "a"   ), '
                    'url(a) format( "123x"  , "a"   )',
                    'local(HiraKakuPro-W3), local(Meiryo), local(IPAPGothic)',
                    'local(Gentium), url(/fonts/Gentium.ttf)',
                    'local("Gentium"), url("/fonts/Gentium.ttf")',
                    'local(Futura-Medium), url(fonts.svg#MyGeometricModern) '
                    'format("svg")',
                ),
            ): (True, True, FM3FF),
            (
                'text-shadow',
                (
                    'none',
                    '1px 1px',
                    '1px 1px 1px',
                    '1px 1px 1px 1px',
                    '1px 1px 1px 1px red',
                    'inset 1px 1px',
                    'inset 1px 1px 1px 1px black',
                ),
            ): (True, True, C3T),
            (
                'text-shadow',
                (
                    '1px',
                    '1px 1px 1px 1px 1px',
                    'x 1px 1px',
                    'inset',
                    '1px black',
                    'black',
                ),
            ): (False, False, C3T),
            ('unicode-range', ('u+1', 'U+111111-ffffff', 'u+123456  ,  U+1-f')): (
                True,
                True,
                FM3FF,
            ),
        }
        # TODO!!!
        for (name, values), (valid, _matching, _profile) in list(tests.items()):
            for value in values:
                assert valid == cssutils.profile.validate(name, value)

                # assert (
                #     valid,
                #     matching,
                #     list(profile),
                # ) == cssutils.profile.validateWithProfile(name, value)

    def _gen_validation_inputs():
        yield (
            (FM3FF, 'font-family', ('y', '"y"')),
            (  # => name should be "y"!!!
                True,
                True,
                FM3FF,
            ),
        )
        yield ((FM3FF, 'font-family', ('"y", "a"', 'a, b', 'a a')), (True, False, CSS2))
        yield (
            (FM3FF, 'font-stretch', ('normal', 'wider', 'narrower', 'inherit')),
            (
                True,
                False,
                FM3,
            ),
        )
        yield ((FM3FF, 'font-style', ('inherit',)), (True, False, CSS2))
        yield (
            (
                FM3FF,
                'font-weight',
                (
                    'bolder',
                    'lighter',
                    'inherit',
                ),
            ),
            (True, False, CSS2),
        )

    @pytest.mark.parametrize(('params', 'results'), _gen_validation_inputs())
    @pytest.mark.xfail(reason="#37")
    def test_validateWithProfile_fonts(self, params, results):
        "Profiles.validateWithProfile()"
        # testing for valid values overwritten in a profile
        v, m, p = results
        expected = v, m, list(p)
        assert cssutils.profile.validateWithProfile(*params) == expected



"""Selector is a single Selector of a CSSStyleRule SelectorList.
Partly implements http://www.w3.org/TR/css3-selectors/.

TODO
    - .contains(selector)
    - .isSubselector(selector)
"""

from __future__ import annotations

__all__ = ['Selector']

import contextlib
import dataclasses
import xml.dom

import cssutils
from cssutils.helper import Deprecated
from cssutils.util import _SimpleNamespaces


class Constants:
    "expected constants"

    # used for equality checks and setting of a space combinator
    S = ' '

    simple_selector_sequence = (
        'type_selector universal HASH class ' 'attrib pseudo negation '
    )
    simple_selector_sequence2 = 'HASH class attrib pseudo negation '

    element_name = 'element_name'

    negation_arg = 'type_selector universal HASH class attrib pseudo'
    negationend = ')'

    attname = 'prefix attribute'
    attname2 = 'attribute'
    attcombinator = 'combinator ]'  # optional
    attvalue = 'value'  # optional
    attend = ']'

    expressionstart = 'PLUS - DIMENSION NUMBER STRING IDENT'
    expression = expressionstart + ' )'

    combinator = ' combinator'


@dataclasses.dataclass
class New(cssutils.util._BaseClass):
    """
    Derives from _BaseClass to provide self._log.
    """

    selector: Selector
    namespaces: dict[str, str]
    context: list[str] = dataclasses.field(default_factory=lambda: [''])
    "stack of: 'attrib', 'negation', 'pseudo'"
    element: str | None = None
    _PREFIX: str | None = None
    specificity: list[int] = dataclasses.field(default_factory=lambda: [0] * 4)
    "mutable, finally a tuple!"
    wellformed: bool = True

    def append(self, seq, val, typ=None, token=None):  # noqa: C901
        """
        appends to seq

        namespace_prefix, IDENT will be combined to a tuple
        (prefix, name) where prefix might be None, the empty string
        or a prefix.

        Saved are also:
            - specificity definition: style, id, class/att, type
            - element: the element this Selector is for
        """
        context = self.context[-1]
        if token:
            line, col = token[2], token[3]
        else:
            line, col = None, None

        if typ == '_PREFIX':
            # SPECIAL TYPE: save prefix for combination with next
            self._PREFIX = val[:-1]
            # handle next time
            return

        if self._PREFIX is not None:
            # as saved from before and reset to None
            prefix, self._PREFIX = self._PREFIX, None
        elif typ == 'universal' and '|' in val:
            # val == *|* or prefix|*
            prefix, val = val.split('|')
        else:
            prefix = None

        # namespace
        if (typ.endswith('-selector') or typ == 'universal') and not (
            'attribute-selector' == typ and not prefix
        ):
            # att **IS NOT** in default ns
            if prefix == '*':
                # *|name: in ANY_NS
                namespaceURI = cssutils._ANYNS
            elif prefix is None:
                # e or *: default namespace with prefix u''
                # or local-name()
                namespaceURI = self.namespaces.get('', None)
            elif prefix == '':
                # |name or |*: in no (or the empty) namespace
                namespaceURI = ''
            else:
                # explicit namespace prefix
                # does not raise KeyError, see _SimpleNamespaces
                namespaceURI = self.namespaces[prefix]

                if namespaceURI is None:
                    self.wellformed = False
                    self._log.error(
                        'Selector: No namespaceURI found ' 'for prefix %r' % prefix,
                        token=token,
                        error=xml.dom.NamespaceErr,
                    )
                    return

            # val is now (namespaceprefix, name) tuple
            val = (namespaceURI, val)

        # specificity
        if not context or context == 'negation':
            if 'id' == typ:
                self.specificity[1] += 1
            elif 'class' == typ or '[' == val:
                self.specificity[2] += 1
            elif typ in (
                'type-selector',
                'negation-type-selector',
                'pseudo-element',
            ):
                self.specificity[3] += 1
        if not context and typ in ('type-selector', 'universal'):
            # define element
            self.element = val

        seq.append(val, typ, line=line, col=col)

    def _COMMENT(self, expected, seq, token, tokenizer=None):
        "special implementation for comment token"
        self.append(seq, cssutils.css.CSSComment([token]), 'COMMENT', token=token)
        return expected

    def _S(self, expected, seq, token, tokenizer=None):
        # S
        context = self.context[-1]
        if context.startswith('pseudo-'):
            if seq and seq[-1].value not in '+-':
                # e.g. x:func(a + b)
                self.append(seq, Constants.S, 'S', token=token)
            return expected

        elif context != 'attrib' and 'combinator' in expected:
            self.append(seq, Constants.S, 'descendant', token=token)
            return Constants.simple_selector_sequence + Constants.combinator

        else:
            return expected

    def _universal(self, expected, seq, token, tokenizer=None):
        # *|* or prefix|*
        context = self.context[-1]
        val = self.selector._tokenvalue(token)
        if 'universal' in expected:
            self.append(seq, val, 'universal', token=token)

            if 'negation' == context:
                return Constants.negationend
            else:
                return Constants.simple_selector_sequence2 + Constants.combinator

        else:
            self.wellformed = False
            self._log.error('Selector: Unexpected universal.', token=token)
            return expected

    def _namespace_prefix(self, expected, seq, token, tokenizer=None):
        # prefix| => element_name
        # or prefix| => attribute_name if attrib
        context = self.context[-1]
        val = self.selector._tokenvalue(token)
        if 'attrib' == context and 'prefix' in expected:
            # [PREFIX|att]
            self.append(seq, val, '_PREFIX', token=token)
            return Constants.attname2
        elif 'type_selector' in expected:
            # PREFIX|*
            self.append(seq, val, '_PREFIX', token=token)
            return Constants.element_name
        else:
            self.wellformed = False
            self._log.error('Selector: Unexpected namespace prefix.', token=token)
            return expected

    def _pseudo(self, expected, seq, token, tokenizer=None):
        # pseudo-class or pseudo-element :a ::a :a( ::a(
        """
        /* '::' starts a pseudo-element, ':' a pseudo-class */
        /* Exceptions: :first-line, :first-letter, :before and
        :after. */
        /* Note that pseudo-elements are restricted to one per selector
        and */
        /* occur only in the last simple_selector_sequence. */
        """
        context = self.context[-1]
        val, typ = (
            self.selector._tokenvalue(token, normalize=True),
            self.selector._type(token),
        )
        if 'pseudo' in expected:
            if val in (':first-line', ':first-letter', ':before', ':after'):
                # always pseudo-element ???
                typ = 'pseudo-element'
            self.append(seq, val, typ, token=token)

            if val.endswith('('):
                # function
                # "pseudo-" "class" or "element"
                self.context.append(typ)
                return Constants.expressionstart
            elif 'negation' == context:
                return Constants.negationend
            elif 'pseudo-element' == typ:
                # only one per element, check at ) also!
                return Constants.combinator
            else:
                return Constants.simple_selector_sequence2 + Constants.combinator

        else:
            self.wellformed = False
            self._log.error('Selector: Unexpected start of pseudo.', token=token)
            return expected

    def _expression(self, expected, seq, token, tokenizer=None):
        # [ [ PLUS | '-' | DIMENSION | NUMBER | STRING | IDENT ] S* ]+
        context = self.context[-1]
        val, typ = self.selector._tokenvalue(token), self.selector._type(token)
        if context.startswith('pseudo-'):
            self.append(seq, val, typ, token=token)
            return Constants.expression
        else:
            self.wellformed = False
            self._log.error('Selector: Unexpected %s.' % typ, token=token)
            return expected

    def _attcombinator(self, expected, seq, token, tokenizer=None):
        # context: attrib
        # PREFIXMATCH | SUFFIXMATCH | SUBSTRINGMATCH | INCLUDES |
        # DASHMATCH
        context = self.context[-1]
        val, typ = self.selector._tokenvalue(token), self.selector._type(token)
        if 'attrib' == context and 'combinator' in expected:
            # combinator in attrib
            self.append(seq, val, typ.lower(), token=token)
            return Constants.attvalue
        else:
            self.wellformed = False
            self._log.error('Selector: Unexpected %s.' % typ, token=token)
            return expected

    def _string(self, expected, seq, token, tokenizer=None):
        # identifier
        context = self.context[-1]
        typ, val = self.selector._type(token), self.selector._stringtokenvalue(token)

        # context: attrib
        if 'attrib' == context and 'value' in expected:
            # attrib: [...=VALUE]
            self.append(seq, val, typ, token=token)
            return Constants.attend

        # context: pseudo
        elif context.startswith('pseudo-'):
            # :func(...)
            self.append(seq, val, typ, token=token)
            return Constants.expression

        else:
            self.wellformed = False
            self._log.error('Selector: Unexpected STRING.', token=token)
            return expected

    def _ident(self, expected, seq, token, tokenizer=None):
        # identifier
        context = self.context[-1]
        val, typ = self.selector._tokenvalue(token), self.selector._type(token)

        # context: attrib
        if 'attrib' == context and 'attribute' in expected:
            # attrib: [...|ATT...]
            self.append(seq, val, 'attribute-selector', token=token)
            return Constants.attcombinator

        elif 'attrib' == context and 'value' in expected:
            # attrib: [...=VALUE]
            self.append(seq, val, 'attribute-value', token=token)
            return Constants.attend

        # context: negation
        elif 'negation' == context:
            # negation: (prefix|IDENT)
            self.append(seq, val, 'negation-type-selector', token=token)
            return Constants.negationend

        # context: pseudo
        elif context.startswith('pseudo-'):
            # :func(...)
            self.append(seq, val, typ, token=token)
            return Constants.expression

        elif 'type_selector' in expected or Constants.element_name == expected:
            # element name after ns or complete type_selector
            self.append(seq, val, 'type-selector', token=token)
            return Constants.simple_selector_sequence2 + Constants.combinator

        else:
            self.wellformed = False
            self._log.error('Selector: Unexpected IDENT.', token=token)
            return expected

    def _class(self, expected, seq, token, tokenizer=None):
        # .IDENT
        context = self.context[-1]
        val = self.selector._tokenvalue(token)
        if 'class' in expected:
            self.append(seq, val, 'class', token=token)

            if 'negation' == context:
                return Constants.negationend
            else:
                return Constants.simple_selector_sequence2 + Constants.combinator

        else:
            self.wellformed = False
            self._log.error('Selector: Unexpected class.', token=token)
            return expected

    def _hash(self, expected, seq, token, tokenizer=None):
        # #IDENT
        context = self.context[-1]
        val = self.selector._tokenvalue(token)
        if 'HASH' in expected:
            self.append(seq, val, 'id', token=token)

            if 'negation' == context:
                return Constants.negationend
            else:
                return Constants.simple_selector_sequence2 + Constants.combinator

        else:
            self.wellformed = False
            self._log.error('Selector: Unexpected HASH.', token=token)
            return expected

    def _char(self, expected, seq, token, tokenizer=None):  # noqa: C901
        # + > ~ ) [ ] + -
        context = self.context[-1]
        val = self.selector._tokenvalue(token)

        # context: attrib
        if ']' == val and 'attrib' == context and ']' in expected:
            # end of attrib
            self.append(seq, val, 'attribute-end', token=token)
            context = self.context.pop()  # attrib is done
            context = self.context[-1]
            if 'negation' == context:
                return Constants.negationend
            else:
                return Constants.simple_selector_sequence2 + Constants.combinator

        if '=' == val and 'attrib' == context and 'combinator' in expected:
            # combinator in attrib
            self.append(seq, val, 'equals', token=token)
            return Constants.attvalue

        # context: negation
        if ')' == val and 'negation' == context and ')' in expected:
            # not(negation_arg)"
            self.append(seq, val, 'negation-end', token=token)
            self.context.pop()  # negation is done
            context = self.context[-1]
            return Constants.simple_selector_sequence + Constants.combinator

        # context: pseudo (at least one expression)
        if val in '+-' and context.startswith('pseudo-'):
            # :func(+ -)"
            _names = {'+': 'plus', '-': 'minus'}
            if val == '+' and seq and seq[-1].value == Constants.S:
                seq.replace(-1, val, _names[val])
            else:
                self.append(seq, val, _names[val], token=token)
            return Constants.expression

        if (
            ')' == val
            and context.startswith('pseudo-')
            and Constants.expression == expected
        ):
            # :func(expression)"
            self.append(seq, val, 'function-end', token=token)
            self.context.pop()  # pseudo is done
            if 'pseudo-element' == context:
                return Constants.combinator
            else:
                return Constants.simple_selector_sequence + Constants.combinator

        # context: ROOT
        if '[' == val and 'attrib' in expected:
            # start of [attrib]
            self.append(seq, val, 'attribute-start', token=token)
            self.context.append('attrib')
            return Constants.attname

        if val in '+>~' and 'combinator' in expected:
            # no other combinator except S may be following
            _names = {
                '>': 'child',
                '+': 'adjacent-sibling',
                '~': 'following-sibling',
            }
            if seq and seq[-1].value == Constants.S:
                seq.replace(-1, val, _names[val])
            else:
                self.append(seq, val, _names[val], token=token)
            return Constants.simple_selector_sequence

        if ',' == val:
            # not a selectorlist
            self.wellformed = False
            self._log.error(
                'Selector: Single selector only.',
                error=xml.dom.InvalidModificationErr,
                token=token,
            )
            return expected

        self.wellformed = False
        self._log.error('Selector: Unexpected CHAR.', token=token)
        return expected

    def _negation(self, expected, seq, token, tokenizer=None):
        # not(
        val = self.selector._tokenvalue(token, normalize=True)
        if 'negation' in expected:
            self.context.append('negation')
            self.append(seq, val, 'negation-start', token=token)
            return Constants.negation_arg
        else:
            self.wellformed = False
            self._log.error('Selector: Unexpected negation.', token=token)
            return expected

    def _atkeyword(self, expected, seq, token, tokenizer=None):
        "invalidates selector"
        self.wellformed = False
        self._log.error('Selector: Unexpected ATKEYWORD.', token=token)
        return expected

    @property
    def productions(self):
        return {
            'CHAR': self._char,
            'class': self._class,
            'HASH': self._hash,
            'STRING': self._string,
            'IDENT': self._ident,
            'namespace_prefix': self._namespace_prefix,
            'negation': self._negation,
            'pseudo-class': self._pseudo,
            'pseudo-element': self._pseudo,
            'universal': self._universal,
            # pseudo
            'NUMBER': self._expression,
            'DIMENSION': self._expression,
            # attribute
            'PREFIXMATCH': self._attcombinator,
            'SUFFIXMATCH': self._attcombinator,
            'SUBSTRINGMATCH': self._attcombinator,
            'DASHMATCH': self._attcombinator,
            'INCLUDES': self._attcombinator,
            'S': self._S,
            'COMMENT': self._COMMENT,
            'ATKEYWORD': self._atkeyword,
        }


class Selector(cssutils.util.Base2):
    """
    (cssutils) a single selector in a :class:`~cssutils.css.SelectorList`
    of a :class:`~cssutils.css.CSSStyleRule`.

    Format::

        # implemented in SelectorList
        selectors_group
          : selector [ COMMA S* selector ]*
          ;

        selector
          : simple_selector_sequence [ combinator simple_selector_sequence ]*
          ;

        combinator
          /* combinators can be surrounded by white space */
          : PLUS S* | GREATER S* | TILDE S* | S+
          ;

        simple_selector_sequence
          : [ type_selector | universal ]
            [ HASH | class | attrib | pseudo | negation ]*
          | [ HASH | class | attrib | pseudo | negation ]+
          ;

        type_selector
          : [ namespace_prefix ]? element_name
          ;

        namespace_prefix
          : [ IDENT | '*' ]? '|'
          ;

        element_name
          : IDENT
          ;

        universal
          : [ namespace_prefix ]? '*'
          ;

        class
          : '.' IDENT
          ;

        attrib
          : '[' S* [ namespace_prefix ]? IDENT S*
                [ [ PREFIXMATCH |
                    SUFFIXMATCH |
                    SUBSTRINGMATCH |
                    '=' |
                    INCLUDES |
                    DASHMATCH ] S* [ IDENT | STRING ] S*
                ]? ']'
          ;

        pseudo
          /* '::' starts a pseudo-element, ':' a pseudo-class */
          /* Exceptions: :first-line, :first-letter, :before and :after. */
          /* Note that pseudo-elements are restricted to one per selector and */
          /* occur only in the last simple_selector_sequence. */
          : ':' ':'? [ IDENT | functional_pseudo ]
          ;

        functional_pseudo
          : FUNCTION S* expression ')'
          ;

        expression
          /* In CSS3, the expressions are identifiers, strings, */
          /* or of the form "an+b" */
          : [ [ PLUS | '-' | DIMENSION | NUMBER | STRING | IDENT ] S* ]+
          ;

        negation
          : NOT S* negation_arg S* ')'
          ;

        negation_arg
          : type_selector | universal | HASH | class | attrib | pseudo
          ;

    """

    def __init__(self, selectorText=None, parent=None, readonly=False):
        """
        :Parameters:
            selectorText
                initial value of this selector
            parent
                a SelectorList
            readonly
                default to False
        """
        super().__init__()

        self.__namespaces = _SimpleNamespaces(log=self._log)
        self._element = None
        self._parent = parent
        self._specificity = (0, 0, 0, 0)

        if selectorText:
            self.selectorText = selectorText

        self._readonly = readonly

    def __repr__(self):
        if self.__getNamespaces():
            st = (self.selectorText, self._getUsedNamespaces())
        else:
            st = self.selectorText
        return f"cssutils.css.{self.__class__.__name__}(selectorText={st!r})"

    def __str__(self):
        return (
            "<cssutils.css.%s object selectorText=%r specificity=%r"
            " _namespaces=%r at 0x%x>"
            % (
                self.__class__.__name__,
                self.selectorText,
                self.specificity,
                self._getUsedNamespaces(),
                id(self),
            )
        )

    def _getUsedUris(self):
        "Return list of actually used URIs in this Selector."
        uris = set()
        for item in self.seq:
            type_, val = item.type, item.value
            if (
                type_.endswith('-selector')
                or type_ == 'universal'
                and isinstance(val, tuple)
                and val[0] not in (None, '*')
            ):
                uris.add(val[0])
        return uris

    def _getUsedNamespaces(self):
        "Return actually used namespaces only."
        useduris = self._getUsedUris()
        namespaces = _SimpleNamespaces(log=self._log)
        for p, uri in list(self._namespaces.items()):
            if uri in useduris:
                namespaces[p] = uri
        return namespaces

    def __getNamespaces(self):
        "Use own namespaces if not attached to a sheet, else the sheet's ones."
        try:
            return self._parent.parentRule.parentStyleSheet.namespaces
        except AttributeError:
            return self.__namespaces

    _namespaces = property(
        __getNamespaces,
        doc="If this Selector is attached to a "
        "CSSStyleSheet the namespaces of that sheet "
        "are mirrored here. While the Selector (or "
        "parent SelectorList or parentRule(s) of that "
        "are not attached a own dict of {prefix: "
        "namespaceURI} is used.",
    )

    @property
    def element(self):
        """Effective element target of this selector."""
        return self._element

    parent = property(
        lambda self: self._parent,
        doc="(DOM) The SelectorList that contains this Selector "
        "or None if this Selector is not attached to a "
        "SelectorList.",
    )

    def _getSelectorText(self):
        """Return serialized format."""
        return cssutils.ser.do_css_Selector(self)

    def _setSelectorText(self, selectorText):
        """
        :param selectorText:
            parsable string or a tuple of (selectorText, dict-of-namespaces).
            Given namespaces are ignored if this object is attached to a
            CSSStyleSheet!

        :exceptions:
            - :exc:`~xml.dom.NamespaceErr`:
              Raised if the specified selector uses an unknown namespace
              prefix.
            - :exc:`~xml.dom.SyntaxErr`:
              Raised if the specified CSS string value has a syntax error
              and is unparsable.
            - :exc:`~xml.dom.NoModificationAllowedErr`:
              Raised if this rule is readonly.
        """
        self._checkReadonly()

        # might be (selectorText, namespaces)
        selectorText, namespaces = self._splitNamespacesOff(selectorText)

        with contextlib.suppress(AttributeError):
            # uses parent stylesheets namespaces if available,
            # otherwise given ones
            namespaces = self.parent.parentRule.parentStyleSheet.namespaces

        tokenizer = self._tokenize2(selectorText)
        if not tokenizer:
            self._log.error('Selector: No selectorText given.')
            return

        tokenizer = self._prepare_tokens(tokenizer)

        new = New(selector=self, namespaces=namespaces)

        # expected: only|not or mediatype, mediatype, feature, and
        newseq = self._tempSeq()

        wellformed, expected = self._parse(
            expected=Constants.simple_selector_sequence,
            seq=newseq,
            tokenizer=tokenizer,
            productions=new.productions,
        )
        wellformed = wellformed and new.wellformed

        # post condition
        if len(new.context) > 1 or not newseq:
            wellformed = False
            self._log.error(
                'Selector: Invalid or incomplete selector: %s'
                % self._valuestr(selectorText)
            )

        if expected == 'element_name':
            wellformed = False
            self._log.error(
                'Selector: No element name found: %s' % self._valuestr(selectorText)
            )

        if expected == Constants.simple_selector_sequence and newseq:
            wellformed = False
            self._log.error(
                'Selector: Cannot end with combinator: %s'
                % self._valuestr(selectorText)
            )

        if (
            newseq
            and hasattr(newseq[-1].value, 'strip')
            and newseq[-1].value.strip() == ''
        ):
            del newseq[-1]

        # set
        if wellformed:
            self.__namespaces = namespaces
            self._element = new.element
            self._specificity = tuple(new.specificity)
            self._setSeq(newseq)
            # filter that only used ones are kept
            self.__namespaces = self._getUsedNamespaces()

    def _prepare_tokens(self, tokenizer):  # noqa: C901
        """
        "*" -> type "universal"
        "*"|IDENT + "|" -> combined to "namespace_prefix"
        "|" -> type "namespace_prefix"
        "." + IDENT -> combined to "class"
        ":" + IDENT, ":" + FUNCTION -> pseudo-class
        FUNCTION "not(" -> negation
        "::" + IDENT, "::" + FUNCTION -> pseudo-element
        """
        tokens = []
        for t in tokenizer:
            typ, val, lin, col = t
            if val == ':' and tokens and self._tokenvalue(tokens[-1]) == ':':
                # combine ":" and ":"
                tokens[-1] = (typ, '::', lin, col)

            elif typ == 'IDENT' and tokens and self._tokenvalue(tokens[-1]) == '.':
                # class: combine to .IDENT
                tokens[-1] = ('class', '.' + val, lin, col)
            elif (
                typ == 'IDENT'
                and tokens
                and self._tokenvalue(tokens[-1]).startswith(':')
                and not self._tokenvalue(tokens[-1]).endswith('(')
            ):
                # pseudo-X: combine to :IDENT or ::IDENT but not ":a(" + "b"
                if self._tokenvalue(tokens[-1]).startswith('::'):
                    t = 'pseudo-element'
                else:
                    t = 'pseudo-class'
                tokens[-1] = (t, self._tokenvalue(tokens[-1]) + val, lin, col)

            elif (
                typ == 'FUNCTION'
                and val == 'not('
                and tokens
                and ':' == self._tokenvalue(tokens[-1])
            ):
                tokens[-1] = ('negation', ':' + val, lin, tokens[-1][3])
            elif (
                typ == 'FUNCTION'
                and tokens
                and self._tokenvalue(tokens[-1]).startswith(':')
            ):
                # pseudo-X: combine to :FUNCTION( or ::FUNCTION(
                if self._tokenvalue(tokens[-1]).startswith('::'):
                    t = 'pseudo-element'
                else:
                    t = 'pseudo-class'
                tokens[-1] = (t, self._tokenvalue(tokens[-1]) + val, lin, col)

            elif (
                val == '*'
                and tokens
                and self._type(tokens[-1]) == 'namespace_prefix'
                and self._tokenvalue(tokens[-1]).endswith('|')
            ):
                # combine prefix|*
                tokens[-1] = (
                    'universal',
                    self._tokenvalue(tokens[-1]) + val,
                    lin,
                    col,
                )
            elif val == '*':
                # universal: "*"
                tokens.append(('universal', val, lin, col))

            elif (
                val == '|'
                and tokens
                and self._type(tokens[-1]) in (self._prods.IDENT, 'universal')
                and self._tokenvalue(tokens[-1]).find('|') == -1
            ):
                # namespace_prefix: "IDENT|" or "*|"
                tokens[-1] = (
                    'namespace_prefix',
                    self._tokenvalue(tokens[-1]) + '|',
                    lin,
                    col,
                )
            elif val == '|':
                # namespace_prefix: "|"
                tokens.append(('namespace_prefix', val, lin, col))

            else:
                tokens.append(t)

        return iter(tokens)

    selectorText = property(
        _getSelectorText,
        _setSelectorText,
        doc="(DOM) The parsable textual representation of " "the selector.",
    )

    specificity = property(
        lambda self: self._specificity,
        doc="""Specificity of this selector (READONLY).
                Tuple of (a, b, c, d) where:

                a
                    presence of style in document, always 0 if not used on a
                    document
                b
                    number of ID selectors
                c
                    number of .class selectors
                d
                    number of Element (type) selectors""",
    )

    wellformed = property(lambda self: bool(len(self.seq)))

    @Deprecated('Use property parent instead')
    def _getParentList(self):
        return self.parent

    parentList = property(_getParentList, doc="DEPRECATED, see property parent instead")



"""Testcases for cssutils.css.CSSCharsetRule"""

import codecs
import os
import sys
import tempfile
from unittest import mock

import pytest

import cssutils

from . import basetest


@pytest.fixture
def serializer(monkeypatch):
    monkeypatch.setattr(cssutils, 'ser', cssutils.serialize.CSSSerializer())


class TestCSSutils(basetest.BaseTestCase):
    exp = '''@import "import/import2.css";
.import {
    /* ./import.css */
    background-image: url(images/example.gif)
    }'''

    def test_parseString(self):
        "cssutils.parseString()"
        s = cssutils.parseString(
            self.exp, media='handheld, screen', title='from string'
        )
        assert isinstance(s, cssutils.css.CSSStyleSheet)
        assert s.href is None
        assert self.exp.encode() == s.cssText
        assert 'utf-8' == s.encoding
        assert 'handheld, screen' == s.media.mediaText
        assert 'from string' == s.title
        assert self.exp.encode() == s.cssText

        ir = s.cssRules[0]
        assert 'import/import2.css' == ir.href
        irs = ir.styleSheet
        assert isinstance(irs, cssutils.css.CSSStyleSheet)

        href = basetest.get_sheet_filename('import.css')
        href = cssutils.helper.path2url(href)
        s = cssutils.parseString(self.exp, href=href)
        assert href == s.href

        ir = s.cssRules[0]
        assert 'import/import2.css' == ir.href
        irs = ir.styleSheet
        assert isinstance(irs, cssutils.css.CSSStyleSheet)
        assert (
            irs.cssText == b'@import "../import3.css";\n'
            b'@import "import-impossible.css" print;\n.import2 {\n'
            b'    /* sheets/import2.css */\n'
            b'    background: url(http://example.com/images/example.gif);\n'
            b'    background: url(//example.com/images/example.gif);\n'
            b'    background: url(/images/example.gif);\n'
            b'    background: url(images2/example.gif);\n'
            b'    background: url(./images2/example.gif);\n'
            b'    background: url(../images/example.gif);\n'
            b'    background: url(./../images/example.gif)\n'
            b'    }'
        )

        tests = {
            'a {color: red}': 'a {\n    color: red\n    }',
            'a {color: rgb(1,2,3)}': 'a {\n    color: rgb(1, 2, 3)\n    }',
        }
        self.do_equal_p(tests)

    def test_parseFile(self, monkeypatch):
        "cssutils.parseFile()"
        # name if used with open, href used for @import resolving
        name = basetest.get_sheet_filename('import.css')
        href = cssutils.helper.path2url(name)

        s = cssutils.parseFile(name, href=href, media='screen', title='from file')
        assert isinstance(s, cssutils.css.CSSStyleSheet)
        if sys.platform.startswith('java'):
            # on Jython only file:
            assert s.href.startswith('file:')
        else:
            # normally file:/// on win and file:/ on unix
            assert s.href.startswith('file:/')
        assert s.href.endswith('/sheets/import.css')
        assert 'utf-8' == s.encoding
        assert 'screen' == s.media.mediaText
        assert 'from file' == s.title
        assert self.exp.encode() == s.cssText

        ir = s.cssRules[0]
        assert 'import/import2.css' == ir.href
        irs = ir.styleSheet
        assert isinstance(irs, cssutils.css.CSSStyleSheet)
        assert (
            irs.cssText == b'@import "../import3.css";\n'
            b'@import "import-impossible.css" print;\n'
            b'.import2 {\n'
            b'    /* sheets/import2.css */\n'
            b'    background: url(http://example.com/images/example.gif);\n'
            b'    background: url(//example.com/images/example.gif);\n'
            b'    background: url(/images/example.gif);\n'
            b'    background: url(images2/example.gif);\n'
            b'    background: url(./images2/example.gif);\n'
            b'    background: url(../images/example.gif);\n'
            b'    background: url(./../images/example.gif)\n'
            b'    }'
        )

        # name is used for open and setting of href automatically
        # test needs to be relative to this test file!
        monkeypatch.chdir(os.path.dirname(__file__))
        name = basetest.get_sheet_filename('import.css')

        s = cssutils.parseFile(name, media='screen', title='from file')
        assert isinstance(s, cssutils.css.CSSStyleSheet)
        if sys.platform.startswith('java'):
            # on Jython only file:
            assert s.href.startswith('file:')
        else:
            # normally file:/// on win and file:/ on unix
            assert s.href.startswith('file:/')
        assert s.href.endswith('/sheets/import.css')
        assert 'utf-8' == s.encoding
        assert 'screen' == s.media.mediaText
        assert 'from file' == s.title
        assert self.exp.encode() == s.cssText

        ir = s.cssRules[0]
        assert 'import/import2.css' == ir.href
        irs = ir.styleSheet
        assert isinstance(irs, cssutils.css.CSSStyleSheet)
        assert (
            irs.cssText == b'@import "../import3.css";\n'
            b'@import "import-impossible.css" print;\n'
            b'.import2 {\n'
            b'    /* sheets/import2.css */\n'
            b'    background: url(http://example.com/images/example.gif);\n'
            b'    background: url(//example.com/images/example.gif);\n'
            b'    background: url(/images/example.gif);\n'
            b'    background: url(images2/example.gif);\n'
            b'    background: url(./images2/example.gif);\n'
            b'    background: url(../images/example.gif);\n'
            b'    background: url(./../images/example.gif)\n'
            b'    }'
        )

        # next test
        css = 'a:after { content: "羊蹄€\u2020" }'

        fd, name = tempfile.mkstemp('_cssutilstest.css')
        t = os.fdopen(fd, 'wb')
        t.write(css.encode('utf-8'))
        t.close()

        with pytest.raises(UnicodeDecodeError):
            cssutils.parseFile(name, 'ascii')

        # ???
        s = cssutils.parseFile(name, encoding='iso-8859-1')
        assert isinstance(s, cssutils.css.CSSStyleSheet)
        assert s.cssRules[1].selectorText == 'a:after'

        s = cssutils.parseFile(name, encoding='utf-8')
        assert isinstance(s, cssutils.css.CSSStyleSheet)
        assert s.cssRules[1].selectorText == 'a:after'

        css = '@charset "iso-8859-1"; a:after { content: "ä" }'
        t = codecs.open(name, 'w', 'iso-8859-1')
        t.write(css)
        t.close()

        with pytest.raises(UnicodeDecodeError):
            cssutils.parseFile(name, 'ascii')

        s = cssutils.parseFile(name, encoding='iso-8859-1')
        assert isinstance(s, cssutils.css.CSSStyleSheet)
        assert s.cssRules[1].selectorText == 'a:after'

        with pytest.raises(UnicodeDecodeError):
            cssutils.parseFile(name, 'utf-8')

        # clean up
        try:
            os.remove(name)
        except OSError:
            pass

    def test_parseUrl(self):
        "cssutils.parseUrl()"
        href = basetest.get_sheet_filename('import.css')
        # href = u'file:' + urllib.pathname2url(href)
        href = cssutils.helper.path2url(href)
        # href = 'http://seewhatever.de/sheets/import.css'
        s = cssutils.parseUrl(href, media='tv, print', title='from url')
        assert isinstance(s, cssutils.css.CSSStyleSheet)
        assert href == s.href
        assert self.exp.encode() == s.cssText
        assert 'utf-8' == s.encoding
        assert 'tv, print' == s.media.mediaText
        assert 'from url' == s.title

        sr = s.cssRules[1]
        img = sr.style.getProperty('background-image').propertyValue[0].value
        assert img == 'images/example.gif'

        ir = s.cssRules[0]
        assert 'import/import2.css' == ir.href
        irs = ir.styleSheet
        assert (
            irs.cssText == b'@import "../import3.css";\n'
            b'@import "import-impossible.css" print;\n'
            b'.import2 {\n'
            b'    /* sheets/import2.css */\n'
            b'    background: url(http://example.com/images/example.gif);\n'
            b'    background: url(//example.com/images/example.gif);\n'
            b'    background: url(/images/example.gif);\n'
            b'    background: url(images2/example.gif);\n'
            b'    background: url(./images2/example.gif);\n'
            b'    background: url(../images/example.gif);\n'
            b'    background: url(./../images/example.gif)\n'
            b'    }'
        )

        ir2 = irs.cssRules[0]
        assert '../import3.css' == ir2.href
        irs2 = ir2.styleSheet
        assert (
            irs2.cssText == b'/* import3 */\n'
            b'.import3 {\n'
            b'    /* from ./import/../import3.css */\n'
            b'    background: url(images/example3.gif);\n'
            b'    background: url(./images/example3.gif);\n'
            b'    background: url(import/images2/example2.gif);\n'
            b'    background: url(./import/images2/example2.gif);\n'
            b'    background: url(import/images2/../../images/example3.gif)\n'
            b'    }'
        )

    def test_setCSSSerializer(self):
        "cssutils.setSerializer() and cssutils.ser"
        s = cssutils.parseString('a { left: 0 }')
        exp4 = '''a {
    left: 0
    }'''
        exp1 = '''a {
 left: 0
 }'''
        assert exp4.encode() == s.cssText
        newser = cssutils.CSSSerializer(cssutils.serialize.Preferences(indent=' '))
        cssutils.setSerializer(newser)
        assert exp1.encode() == s.cssText
        newser = cssutils.CSSSerializer(cssutils.serialize.Preferences(indent='    '))
        cssutils.ser = newser
        assert exp4.encode() == s.cssText

    def test_parseStyle(self):
        "cssutils.parseStyle()"
        s = cssutils.parseStyle('x:0; y:red')
        assert isinstance(s, cssutils.css.CSSStyleDeclaration)
        assert s.cssText == 'x: 0;\ny: red'

        s = cssutils.parseStyle('@import "x";')
        assert isinstance(s, cssutils.css.CSSStyleDeclaration)
        assert s.cssText == ''

        tests = [('content: "ä"', 'iso-8859-1'), ('content: "€"', 'utf-8')]
        for v, e in tests:
            s = cssutils.parseStyle(v.encode(e), encoding=e)
            assert s.cssText == v

        with pytest.raises(UnicodeDecodeError):
            cssutils.parseStyle(
                'content: "ä"'.encode(),
                'ascii',
            )

    def test_getUrls(self):
        "cssutils.getUrls()"
        cssutils.ser.prefs.keepAllProperties = True

        css = r'''
        @import "im1";
        @import url(im2);
        @import url( im3 );
        @import url( "im4" );
        @import url( 'im5' );
        a {
            background-image: url(a) !important;
            background-\image: url(b);
            background: url(c) no-repeat !important;
            /* issue #46 */
            src: local("xx"),
                 url("f.woff") format("woff"),
                 url("f.otf") format("opentype"),
                 url("f.svg#f") format("svg");
            }'''
        urls = set(cssutils.getUrls(cssutils.parseString(css)))
        assert urls == {
            "im1",
            "im2",
            "im3",
            "im4",
            "im5",
            "a",
            "b",
            "c",
            'f.woff',
            'f.svg#f',
            'f.otf',
        }
        cssutils.ser.prefs.keepAllProperties = False

    def test_replaceUrls(self):
        "cssutils.replaceUrls()"
        cssutils.ser.prefs.keepAllProperties = True

        css = r'''
        @import "im1";
        @import url(im2);
        a {
            background-image: url(c) !important;
            background-\image: url(b);
            background: url(a) no-repeat !important;
            }'''
        s = cssutils.parseString(css)
        cssutils.replaceUrls(s, lambda old: "NEW" + old)
        assert '@import "NEWim1";' == s.cssRules[0].cssText
        assert 'NEWim2' == s.cssRules[1].href
        assert (
            '''background-image: url(NEWc) !important;
background-\\image: url(NEWb);
background: url(NEWa) no-repeat !important'''
            == s.cssRules[2].style.cssText
        )

        cssutils.ser.prefs.keepAllProperties = False

        # CSSStyleDeclaration
        style = cssutils.parseStyle(
            '''color: red;
                                        background-image:
                                            url(1.png),
                                            url('2.png')'''
        )
        cssutils.replaceUrls(style, lambda url: 'prefix/' + url)
        assert (
            style.cssText
            == '''color: red;
background-image: url(prefix/1.png), url(prefix/2.png)'''
        )

    @pytest.mark.usefixtures('serializer')
    def test_resolveImports(self):
        "cssutils.resolveImports(sheet)"
        cssutils.ser.prefs.useMinified()

        a = '@charset "iso-8859-1";@import"b.css";\xe4{color:green}'.encode(
            'iso-8859-1'
        )
        b = b'@charset "ascii";\\E4 {color:red}'

        # normal
        m = mock.Mock()
        with mock.patch('cssutils.util._defaultFetcher', m):
            m.return_value = (None, b)
            s = cssutils.parseString(a)

            # py3 TODO
            assert a == s.cssText
            assert b == s.cssRules[1].styleSheet.cssText

            c = cssutils.resolveImports(s)

            # py3 TODO
            assert b'\xc3\xa4{color:red}\xc3\xa4{color:green}' == c.cssText

            c.encoding = 'ascii'
            assert rb'@charset "ascii";\E4 {color:red}\E4 {color:green}' == c.cssText

        # b cannot be found
        m = mock.Mock()
        with mock.patch('cssutils.util._defaultFetcher', m):
            m.return_value = (None, None)
            s = cssutils.parseString(a)

            # py3 TODO
            assert a == s.cssText
            assert isinstance(s.cssRules[1].styleSheet, cssutils.css.CSSStyleSheet)
            c = cssutils.resolveImports(s)
            # py3 TODO
            assert b'@import"b.css";\xc3\xa4{color:green}' == c.cssText

        # @import with media
        a = '@import"b.css";@import"b.css" print, tv ;@import"b.css" all;'
        b = 'a {color: red}'
        m = mock.Mock()
        with mock.patch('cssutils.util._defaultFetcher', m):
            m.return_value = (None, b)
            s = cssutils.parseString(a)

            c = cssutils.resolveImports(s)

            assert b'a{color:red}@media print,tv{a{color:red}}a{color:red}' == c.cssText

        # cannot resolve with media => keep original
        a = '@import"b.css"print;'
        b = '@namespace "http://example.com";'
        m = mock.Mock()
        with mock.patch('cssutils.util._defaultFetcher', m):
            m.return_value = (None, b)
            s = cssutils.parseString(a)
            c = cssutils.resolveImports(s)
            assert a.encode() == c.cssText

        # urls are adjusted too, layout:
        # a.css
        # c.css
        # img/img.gif
        # b/
        #     b.css
        #     subimg/subimg.gif
        a = '''
             @import"b/b.css";
             a {
                 x: url(/img/abs.gif);
                 y: url(img/img.gif);
                 z: url(b/subimg/subimg.gif);
                 }'''

        def fetcher(url):
            c = {
                'b.css': '''
                     @import"../c.css";
                     b {
                         x: url(/img/abs.gif);
                         y: url(../img/img.gif);
                         z: url(subimg/subimg.gif);
                         }''',
                'c.css': '''
                     c {
                         x: url(/img/abs.gif);
                         y: url(./img/img.gif);
                         z: url(./b/subimg/subimg.gif);
                         }''',
            }
            return 'utf-8', c[os.path.split(url)[1]]

        @mock.patch.object(cssutils.util, '_defaultFetcher', new=fetcher)
        def do():
            s = cssutils.parseString(a)
            r = cssutils.resolveImports(s)
            return s, r

        s, r = do()

        cssutils.ser.prefs.useDefaults()
        cssutils.ser.prefs.keepComments = False
        expected = b'''c {
    x: url(/img/abs.gif);
    y: url(img/img.gif);
    z: url(b/subimg/subimg.gif)
    }
b {
    x: url(/img/abs.gif);
    y: url(img/img.gif);
    z: url(b/subimg/subimg.gif)
    }
a {
    x: url(/img/abs.gif);
    y: url(img/img.gif);
    z: url(b/subimg/subimg.gif)
    }'''
        assert expected == r.cssText
